"""
OpenColorIO Processing
======================

Defines the object for *OpenColorIO* processing:

-   :func:`colour.io.process_image_OpenColorIO`
"""

from __future__ import annotations

import numpy as np

from colour.hints import Any, ArrayLike, NDArray
from colour.utilities import as_float_array, required

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "process_image_OpenColorIO",
]


@required("OpenColorIO")
def process_image_OpenColorIO(
    a: ArrayLike, *args: Any, **kwargs: Any
) -> NDArray:
    """
    Process given image with *OpenColorIO*.

    Parameters
    ----------
    a
        Image to process with *OpenColorIO*.

    Other Parameters
    ----------------
    config
        *OpenColorIO* config to use for processing. If not defined, the
        *OpenColorIO* set defined by the ``$OCIO`` environment variable is
        used.
    args
        Arguments for `Config.getProcessor` method.
        See https://opencolorio.readthedocs.io/en/latest/api/config.html for
        more information.

    Returns
    -------
    :class:`numpy.ndarray`
        Processed image.

    Examples
    --------
    # TODO: Reinstate when "Pypi" wheel compatible with "ARM" on "macOS" is
    # released.

    >>> import os
    >>> import PyOpenColorIO as ocio  # doctest: +SKIP
    >>> from colour.utilities import full
    >>> config = os.path.join(
    ...     os.path.dirname(__file__), 'tests', 'resources',
    ...     'config-aces-reference.ocio.yaml')
    >>> a = full([4, 2, 3], 0.18)
    >>> process_image_OpenColorIO(  # doctest: +SKIP
    ...     a, 'ACES - ACES2065-1', 'ACES - ACEScct', config=config)
    array([[[ 0.4135878...,  0.4135878...,  0.4135878...],
            [ 0.4135878...,  0.4135878...,  0.4135878...]],
    <BLANKLINE>
           [[ 0.4135878...,  0.4135878...,  0.4135878...],
            [ 0.4135878...,  0.4135878...,  0.4135878...]],
    <BLANKLINE>
           [[ 0.4135878...,  0.4135878...,  0.4135878...],
            [ 0.4135878...,  0.4135878...,  0.4135878...]],
    <BLANKLINE>
           [[ 0.4135878...,  0.4135878...,  0.4135878...],
            [ 0.4135878...,  0.4135878...,  0.4135878...]]], dtype=float32)
    >>> process_image_OpenColorIO(  # doctest: +SKIP
    ...     a, 'ACES - ACES2065-1', 'Display - sRGB',
    ...     'Output - SDR Video - ACES 1.0', ocio.TRANSFORM_DIR_FORWARD,
    ...     config=config)
    array([[[ 0.3559523...,  0.3559525...,  0.3559525...],
            [ 0.3559523...,  0.3559525...,  0.3559525...]],
    <BLANKLINE>
           [[ 0.3559523...,  0.3559525...,  0.3559525...],
            [ 0.3559523...,  0.3559525...,  0.3559525...]],
    <BLANKLINE>
           [[ 0.3559523...,  0.3559525...,  0.3559525...],
            [ 0.3559523...,  0.3559525...,  0.3559525...]],
    <BLANKLINE>
           [[ 0.3559523...,  0.3559525...,  0.3559525...],
            [ 0.3559523...,  0.3559525...,  0.3559525...]]], dtype=float32)
    """

    import PyOpenColorIO as ocio

    config = kwargs.get("config")
    config = (
        ocio.Config.CreateFromEnv()
        if config is None
        else ocio.Config.CreateFromFile(config)
    )

    a = as_float_array(np.atleast_3d(a), dtype=np.float32)
    height, width, channels = a.shape

    processor = config.getProcessor(*args).getDefaultCPUProcessor()

    image_desc = ocio.PackedImageDesc(a, width, height, channels)

    processor.apply(image_desc)

    return image_desc.getData().reshape([height, width, channels])



"""Defines the unit tests for the :mod:`colour.models.rgb.datasets` module."""

import numpy as np
import pickle
import unittest
from copy import deepcopy

from colour.models import (
    RGB_COLOURSPACES,
    normalised_primary_matrix,
)
from colour.utilities import as_int, ignore_numpy_errors

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestRGB_COLOURSPACES",
]


class TestRGB_COLOURSPACES(unittest.TestCase):
    """
    Define :attr:`colour.models.rgb.datasets.RGB_COLOURSPACES`
    attribute unit tests methods.
    """

    def test_transformation_matrices(self):
        """
        Test the transformations matrices from the
        :attr:`colour.models.rgb.datasets.RGB_COLOURSPACES` attribute
        colourspace models.
        """

        tolerances = {
            "Adobe RGB (1998)": 1e-5,
            "ALEXA Wide Gamut": 1e-6,
            "DJI D-Gamut": 1e-4,
            "ERIMM RGB": 1e-3,
            "ProPhoto RGB": 1e-3,
            "REDWideGamutRGB": 1e-6,
            "RIMM RGB": 1e-3,
            "ROMM RGB": 1e-3,
            "sRGB": 1e-4,
            "V-Gamut": 1e-6,
        }
        XYZ_r = np.array([0.5, 0.5, 0.5]).reshape([3, 1])
        for colourspace in RGB_COLOURSPACES.values():
            M = normalised_primary_matrix(
                colourspace.primaries, colourspace.whitepoint
            )

            tolerance = tolerances.get(colourspace.name, 1e-7)
            np.testing.assert_allclose(
                colourspace.matrix_RGB_to_XYZ,
                M,
                rtol=tolerance,
                atol=tolerance,
                verbose=False,
            )

            RGB = np.dot(colourspace.matrix_XYZ_to_RGB, XYZ_r)
            XYZ = np.dot(colourspace.matrix_RGB_to_XYZ, RGB)
            np.testing.assert_allclose(
                XYZ_r, XYZ, rtol=tolerance, atol=tolerance, verbose=False
            )

            # Derived transformation matrices.
            colourspace = deepcopy(colourspace)
            colourspace.use_derived_transformation_matrices(True)
            RGB = np.dot(colourspace.matrix_XYZ_to_RGB, XYZ_r)
            XYZ = np.dot(colourspace.matrix_RGB_to_XYZ, RGB)
            np.testing.assert_almost_equal(XYZ_r, XYZ, decimal=7)

    def test_cctf(self):
        """
        Test colour component transfer functions from the
        :attr:`colour.models.rgb.datasets.RGB_COLOURSPACES` attribute
        colourspace models.
        """

        ignored_colourspaces = ("ACESproxy",)

        decimals = {"DJI D-Gamut": 1, "F-Gamut": 4, "N-Gamut": 3}

        samples = np.hstack(
            [np.linspace(0, 1, as_int(1e5)), np.linspace(0, 65504, 65504 * 10)]
        )

        for colourspace in RGB_COLOURSPACES.values():
            if colourspace.name in ignored_colourspaces:
                continue

            cctf_encoding_s = colourspace.cctf_encoding(samples)
            cctf_decoding_s = colourspace.cctf_decoding(cctf_encoding_s)

            np.testing.assert_almost_equal(
                samples,
                cctf_decoding_s,
                decimal=decimals.get(colourspace.name, 7),
            )

    def test_n_dimensional_cctf(self):
        """
        Test colour component transfer functions from the
        :attr:`colour.models.rgb.datasets.RGB_COLOURSPACES` attribute
        colourspace models n-dimensional arrays support.
        """

        decimals = {"DJI D-Gamut": 6, "F-Gamut": 4}

        for colourspace in RGB_COLOURSPACES.values():
            value_cctf_encoding = 0.5
            value_cctf_decoding = colourspace.cctf_decoding(
                colourspace.cctf_encoding(value_cctf_encoding)
            )
            np.testing.assert_almost_equal(
                value_cctf_encoding,
                value_cctf_decoding,
                decimal=decimals.get(colourspace.name, 7),
            )

            value_cctf_encoding = np.tile(value_cctf_encoding, 6)
            value_cctf_decoding = np.tile(value_cctf_decoding, 6)
            np.testing.assert_almost_equal(
                value_cctf_encoding,
                value_cctf_decoding,
                decimal=decimals.get(colourspace.name, 7),
            )

            value_cctf_encoding = np.reshape(value_cctf_encoding, (3, 2))
            value_cctf_decoding = np.reshape(value_cctf_decoding, (3, 2))
            np.testing.assert_almost_equal(
                value_cctf_encoding,
                value_cctf_decoding,
                decimal=decimals.get(colourspace.name, 7),
            )

            value_cctf_encoding = np.reshape(value_cctf_encoding, (3, 2, 1))
            value_cctf_decoding = np.reshape(value_cctf_decoding, (3, 2, 1))
            np.testing.assert_almost_equal(
                value_cctf_encoding,
                value_cctf_decoding,
                decimal=decimals.get(colourspace.name, 7),
            )

    @ignore_numpy_errors
    def test_nan_cctf(self):
        """
        Test colour component transfer functions from the
        :attr:`colour.models.rgb.datasets.RGB_COLOURSPACES` attribute
        colourspace models nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        for colourspace in RGB_COLOURSPACES.values():
            for case in cases:
                colourspace.cctf_encoding(case)
                colourspace.cctf_decoding(case)

    def test_pickle(self):
        """Test the "pickle-ability" of the *RGB* colourspaces."""

        for colourspace in RGB_COLOURSPACES.values():
            pickle.dumps(colourspace)


if __name__ == "__main__":
    unittest.main()



"""Showcases input / output *CSV* tabular data related examples."""

import os
from pprint import pprint

import colour
from colour.utilities import message_box

RESOURCES_DIRECTORY = os.path.join(os.path.dirname(__file__), "resources")

message_box('"CSV" Tabular Data IO')

message_box('Reading tabular data from "CSV" file.')
data_babelcolor_average = colour.read_spectral_data_from_csv_file(
    os.path.join(RESOURCES_DIRECTORY, "babelcolor_average.csv")
)
pprint(sorted(data_babelcolor_average.keys()))

print("\n")

message_box(
    'Reading spectral data from a "CSV" file directly as spectral '
    "distributions."
)
sds_babelcolor_average = colour.read_sds_from_csv_file(
    os.path.join(RESOURCES_DIRECTORY, "babelcolor_average.csv")
)
pprint(sds_babelcolor_average)



from .datasets import *  # noqa
from . import datasets
from .macadam_limits import is_within_macadam_limits
from .mesh import is_within_mesh_volume
from .pointer_gamut import is_within_pointer_gamut
from .spectrum import (
    generate_pulse_waves,
    XYZ_outer_surface,
    solid_RoschMacAdam,
    is_within_visible_spectrum,
)
from .rgb import (
    RGB_colourspace_limits,
    RGB_colourspace_volume_MonteCarlo,
    RGB_colourspace_volume_coverage_MonteCarlo,
    RGB_colourspace_pointer_gamut_coverage_MonteCarlo,
    RGB_colourspace_visible_spectrum_coverage_MonteCarlo,
)

__all__ = []
__all__ += datasets.__all__
__all__ += [
    "is_within_macadam_limits",
]
__all__ += [
    "is_within_mesh_volume",
]
__all__ += [
    "is_within_pointer_gamut",
]
__all__ += [
    "generate_pulse_waves",
    "XYZ_outer_surface",
    "solid_RoschMacAdam",
    "is_within_visible_spectrum",
]
__all__ += [
    "RGB_colourspace_limits",
    "RGB_colourspace_volume_MonteCarlo",
    "RGB_colourspace_volume_coverage_MonteCarlo",
    "RGB_colourspace_pointer_gamut_coverage_MonteCarlo",
    "RGB_colourspace_visible_spectrum_coverage_MonteCarlo",
]



import sys

from colour.utilities.deprecation import ModuleAPI, build_API_changes
from colour.utilities.documentation import is_documentation_building

from colour.hints import Any

from .michaelis_menten import (
    REACTION_RATE_MICHAELISMENTEN_METHODS,
    reaction_rate_MichaelisMenten,
    SUBSTRATE_CONCENTRATION_MICHAELISMENTEN_METHODS,
    substrate_concentration_MichaelisMenten,
)
from .michaelis_menten import (
    reaction_rate_MichaelisMenten_Michaelis1913,
    substrate_concentration_MichaelisMenten_Michaelis1913,
    reaction_rate_MichaelisMenten_Abebe2017,
    substrate_concentration_MichaelisMenten_Abebe2017,
)

__all__ = [
    "REACTION_RATE_MICHAELISMENTEN_METHODS",
    "reaction_rate_MichaelisMenten",
    "SUBSTRATE_CONCENTRATION_MICHAELISMENTEN_METHODS",
    "substrate_concentration_MichaelisMenten",
]
__all__ += [
    "reaction_rate_MichaelisMenten_Michaelis1913",
    "substrate_concentration_MichaelisMenten_Michaelis1913",
    "reaction_rate_MichaelisMenten_Abebe2017",
    "substrate_concentration_MichaelisMenten_Abebe2017",
]


# ----------------------------------------------------------------------------#
# ---                API Changes and Deprecation Management                ---#
# ----------------------------------------------------------------------------#
class biochemistry(ModuleAPI):
    """Define a class acting like the *biochemistry* module."""

    def __getattr__(self, attribute) -> Any:
        """Return the value from the attribute with given name."""

        return super().__getattr__(attribute)


# v0.4.0
API_CHANGES = {
    "ObjectRenamed": [
        [
            "colour.biochemistry.reaction_rate_MichealisMenten",
            "colour.biochemistry.reaction_rate_MichaelisMenten",
        ],
        [
            "colour.biochemistry.substrate_concentration_MichealisMenten",
            "colour.biochemistry.substrate_concentration_MichaelisMenten",
        ],
    ]
}
"""Defines the *colour.biochemistry* sub-package API changes."""

if not is_documentation_building():
    sys.modules[
        "colour.biochemistry"
    ] = biochemistry(  # type:ignore[assignment]
        sys.modules["colour.biochemistry"], build_API_changes(API_CHANGES)
    )

    del ModuleAPI, is_documentation_building, build_API_changes, sys



"""
ITU-R BT.2020 Colourspace
=========================

Defines the *ITU-R BT.2020* colourspace:

-   :attr:`colour.models.RGB_COLOURSPACE_BT2020`.

References
----------
-   :cite:`InternationalTelecommunicationUnion2015h` : International
    Telecommunication Union. (2015). Recommendation ITU-R BT.2020 - Parameter
    values for ultra-high definition television systems for production and
    international programme exchange (pp. 1-8).
    https://www.itu.int/dms_pubrec/itu-r/rec/bt/\
R-REC-BT.2020-2-201510-I!!PDF-E.pdf
"""

from __future__ import annotations

import numpy as np

from colour.colorimetry import CCS_ILLUMINANTS
from colour.hints import NDArray
from colour.models.rgb import (
    RGB_Colourspace,
    normalised_primary_matrix,
    eotf_inverse_BT2020,
    eotf_BT2020,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "PRIMARIES_BT2020",
    "WHITEPOINT_NAME_BT2020",
    "CCS_WHITEPOINT_BT2020",
    "MATRIX_BT2020_TO_XYZ",
    "MATRIX_XYZ_TO_BT2020",
    "RGB_COLOURSPACE_BT2020",
]

PRIMARIES_BT2020: NDArray = np.array(
    [
        [0.7080, 0.2920],
        [0.1700, 0.7970],
        [0.1310, 0.0460],
    ]
)
"""*ITU-R BT.2020* colourspace primaries."""

WHITEPOINT_NAME_BT2020: str = "D65"
"""*ITU-R BT.2020* colourspace whitepoint name."""

CCS_WHITEPOINT_BT2020: NDArray = CCS_ILLUMINANTS[
    "CIE 1931 2 Degree Standard Observer"
][WHITEPOINT_NAME_BT2020]
"""*ITU-R BT.2020* colourspace whitepoint chromaticity coordinates."""

MATRIX_BT2020_TO_XYZ: NDArray = normalised_primary_matrix(
    PRIMARIES_BT2020, CCS_WHITEPOINT_BT2020
)
"""*ITU-R BT.2020* colourspace to *CIE XYZ* tristimulus values matrix."""

MATRIX_XYZ_TO_BT2020: NDArray = np.linalg.inv(MATRIX_BT2020_TO_XYZ)
"""*CIE XYZ* tristimulus values to *ITU-R BT.2020* colourspace matrix."""

RGB_COLOURSPACE_BT2020: RGB_Colourspace = RGB_Colourspace(
    "ITU-R BT.2020",
    PRIMARIES_BT2020,
    CCS_WHITEPOINT_BT2020,
    WHITEPOINT_NAME_BT2020,
    MATRIX_BT2020_TO_XYZ,
    MATRIX_XYZ_TO_BT2020,
    eotf_inverse_BT2020,
    eotf_BT2020,
)
RGB_COLOURSPACE_BT2020.__doc__ = """
*ITU-R BT.2020* colourspace.

The wavelength of the *ITU-R BT.2020* primary colours are:

-   630nm for the red primary colour
-   532nm for the green primary colour
-   467nm for the blue primary colour.

References
----------
:cite:`InternationalTelecommunicationUnion2015h`
"""



"""Showcases colour models computations."""

import numpy as np

import colour
from colour.utilities import message_box

message_box("Colour Models Computations")

XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
message_box(
    f'Converting to the "CIE xyY" colourspace from given "CIE XYZ" '
    f"tristimulus values:\n\n\t{XYZ}"
)
print(colour.XYZ_to_xyY(XYZ))

print("\n")

message_box(
    "The default illuminant if X == Y == Z == 0 is "
    '"CIE Standard Illuminant D Series D65".'
)
print(colour.XYZ_to_xyY(np.array([0.00000000, 0.00000000, 0.00000000])))

print("\n")

message_box("Using an alternative illuminant.")
print(
    colour.XYZ_to_xyY(
        np.array([0.00000000, 0.00000000, 0.00000000]),
        colour.CCS_ILLUMINANTS["CIE 1931 2 Degree Standard Observer"]["ACES"],
    )
)

print("\n")

xyY = np.array([0.26414772, 0.37770001, 0.10080000])
message_box(
    f'Converting to "CIE XYZ" tristimulus values from given "CIE xyY" '
    f"colourspace values:\n\n\t{xyY}"
)
print(colour.xyY_to_XYZ(xyY))

print("\n")

message_box(
    f'Converting to "xy" chromaticity coordinates from given "CIE XYZ" '
    f"tristimulus values:\n\n\t{XYZ}"
)
print(colour.XYZ_to_xy(XYZ))

print("\n")

xy = np.array([0.26414772, 0.37770001])
message_box(
    f'Converting to "CIE XYZ" tristimulus values from given "xy" chromaticity '
    f"coordinates:\n\n\t{xy}"
)
print(colour.xy_to_XYZ(xy))

print("\n")

message_box(
    f'Converting to "RGB" colourspace from given "CIE XYZ" tristimulus '
    f"values:\n\n\t{XYZ}"
)
D65 = colour.CCS_ILLUMINANTS["CIE 1931 2 Degree Standard Observer"]["D65"]
print(
    colour.XYZ_to_RGB(
        XYZ,
        D65,
        colour.RGB_COLOURSPACES["sRGB"].whitepoint,
        colour.RGB_COLOURSPACES["sRGB"].matrix_XYZ_to_RGB,
        "Bradford",
        colour.RGB_COLOURSPACES["sRGB"].cctf_encoding,
    )
)

print("\n")

RGB = np.array([0.45620519, 0.03081071, 0.04091952])
message_box(
    f'Converting to "CIE XYZ" tristimulus values from given "RGB" colourspace '
    f"values:\n\n\t{RGB}"
)
print(
    colour.RGB_to_XYZ(
        RGB,
        colour.RGB_COLOURSPACES["sRGB"].whitepoint,
        D65,
        colour.RGB_COLOURSPACES["sRGB"].matrix_RGB_to_XYZ,
        "Bradford",
        colour.RGB_COLOURSPACES["sRGB"].cctf_decoding,
    )
)

print("\n")

message_box(
    f'Converting to "sRGB" colourspace from given "CIE XYZ" tristimulus '
    f"values using convenient definition:\n\n\t{XYZ}"
)
print(colour.XYZ_to_sRGB(XYZ, D65))

print("\n")

message_box(
    f'Converting to "CIE 1960 UCS" colourspace from given "CIE XYZ" '
    f"tristimulus values:\n\n\t{XYZ}"
)
print(colour.XYZ_to_UCS(XYZ))

print("\n")

UCS = np.array([0.07049533, 0.10080000, 0.09558313])
message_box(
    f'Converting to "CIE XYZ" tristimulus values from given "CIE 1960 UCS" '
    f"colourspace values:\n\n\t{UCS}"
)
print(colour.UCS_to_XYZ(UCS))

print("\n")

message_box(
    f'Converting to "uv" chromaticity coordinates from given "CIE UCS" '
    f"colourspace values:\n\n\t{UCS}"
)
print(colour.UCS_to_uv(UCS))

print("\n")

uv = np.array([0.15085309, 0.32355314])
message_box(
    f'Converting to "xy" chromaticity coordinates from given "CIE UCS" '
    f'colourspace "uv" chromaticity coordinates:\n\n\t{uv}'
)
print(colour.UCS_uv_to_xy(uv))

print("\n")

xy = np.array([0.26414771, 0.37770001])
message_box(
    f'Converting to "CIE UCS" colourspace "uv" chromaticity coordinates from '
    f'given "xy" chromaticity coordinates:\n\n\t{xy}'
)
print(colour.xy_to_UCS_uv(xy))

print("\n")

message_box(
    f'Converting to "CIE 1964 U*V*W*" colourspace from given"CIE XYZ" '
    f"tristimulus values:\n\n\t{XYZ * 100}"
)
print(colour.XYZ_to_UVW(XYZ * 100))

print("\n")

UVW = np.array([-22.59840563, 5.45505477, 37.00411491])
message_box(
    f'Converting to "CIE XYZ" tristimulus values from given"CIE 1964 U*V*W*" '
    f"colourspace values:\n\n\t{UVW}"
)
print(colour.UVW_to_XYZ(UVW) / 100)

print("\n")

message_box(
    f'Converting to "CIE L*u*v*" colourspace from given "CIE XYZ" tristimulus '
    f"values:\n\n\t{XYZ}"
)
print(colour.XYZ_to_Luv(XYZ))

print("\n")

Luv = np.array([37.9856291, -23.19781615, 8.39962073])
message_box(
    f'Converting to "CIE XYZ" tristimulus values from given "CIE L*u*v*" '
    f"colourspace values:\n\n\t{Luv}"
)
print(colour.Luv_to_XYZ(Luv))

print("\n")

message_box(
    f'Converting to "u"v"" chromaticity coordinates from given "CIE L*u*v*" '
    f"colourspace values:\n\n\t{Luv}"
)
print(colour.Luv_to_uv(Luv))

print("\n")

uv = np.array([0.1508531, 0.48532971])
message_box(
    f'Converting to "xy" chromaticity coordinates from given "CIE L*u*v*" '
    f'colourspace "u"v"" chromaticity coordinates:\n\n\t{uv}'
)
print(colour.Luv_uv_to_xy(uv))

print("\n")

xy = np.array([0.26414771, 0.37770001])
message_box(
    f'Converting to "CIE L*u*v*" colourspace "u"v"" chromaticity coordinates '
    f'from given "xy" chromaticity coordinates:\n\n\t{xy}'
)
print(colour.xy_to_Luv_uv(xy))

print("\n")

message_box(
    f'Converting to "CIE L*C*Huv" colourspace from given "CIE L*u*v*" '
    f"colourspace values:\n\n\t{Luv}"
)
print(colour.Luv_to_LCHuv(Luv))

print("\n")

LCHuv = np.array([37.9856291, 24.67169031, 160.09535205])
message_box(
    f'Converting to "CIE L*u*v*" colourspace from given "CIE L*C*Huv" '
    f"colourspace values:\n\n\t{LCHuv}"
)
print(colour.LCHuv_to_Luv(LCHuv))

print("\n")

message_box(
    f'Converting to "CIE L*a*b*" colourspace from given "CIE XYZ" tristimulus '
    f"values:\n\n\t{XYZ}"
)
print(colour.XYZ_to_Lab(XYZ))

print("\n")

Lab = np.array([37.9856291, -22.61920654, 4.19811236])
message_box(
    f'Converting to "CIE XYZ" tristimulus values from given "CIE L*a*b*" '
    f"colourspace values:\n\n\t{Lab}"
)
print(colour.Lab_to_XYZ(Lab))

print("\n")

message_box(
    f'Converting to "CIE L*C*Hab" colourspace from given "CIE L*a*b*" '
    f"colourspace values:\n\n\t{Lab}"
)
print(colour.Lab_to_LCHab(Lab))

print("\n")

LCHab = np.array([37.9856291, 23.00549178, 169.48557589])
message_box(
    f'Converting to "CIE L*a*b*" colourspace from given "CIE L*C*Hab" '
    f"colourspace values:\n\n\t{LCHab}"
)
print(colour.LCHab_to_Lab(LCHab))

print("\n")

XYZ = np.array([0.20654008, 0.12197225, 0.05136952]) * 100
message_box(
    f'Converting to "Hunter L,a,b" colour scale from given "CIE XYZ" '
    f"tristimulus values:\n\n\t{XYZ}"
)
print(colour.XYZ_to_Hunter_Lab(XYZ))

print("\n")

Lab = np.array([31.74901573, -14.44108591, 2.74396261])
message_box(
    f'Converting to "CIE XYZ" tristimulus values from given "Hunter L,a,b" '
    f"colour scale values:\n\n\t{Lab}"
)
print(colour.Hunter_Lab_to_XYZ(Lab))

print("\n")

message_box(
    f'Converting to "Hunter Rd,a,b" colour scale from given "CIE XYZ" '
    f"tristimulus values:\n\n\t{XYZ}"
)
print(colour.XYZ_to_Hunter_Rdab(XYZ))

print("\n")

R_d_ab = np.array([10.08000000, -17.8442708, 3.39060457])
message_box(
    f'Converting to "CIE XYZ" tristimulus values from given"Hunter Rd,a,b" '
    f"colour scale values:\n\n\t{R_d_ab}"
)
print(colour.Hunter_Rdab_to_XYZ(R_d_ab))

print("\n")

XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
message_box(
    f'Converting to "ICaCb" colourspace from given "CIE XYZ" tristimulus '
    f"values:\n\n\t{XYZ}"
)
print(colour.XYZ_to_ICaCb(XYZ))

ICaCb = np.array([0.06875297, 0.05753352, 0.02081548])
message_box(
    f'Converting to "CIE XYZ" tristimulus values from given "ICaCb" '
    f"colourspace values:\n\n\t{ICaCb}"
)
print(colour.ICaCb_to_XYZ(ICaCb))

XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
message_box(
    f'Converting to "IgPgTg" colourspace from given "CIE XYZ" tristimulus '
    f"values:\n\n\t{XYZ}"
)
print(colour.XYZ_to_IgPgTg(XYZ))

print("\n")

IgPgTg = np.array([0.42421258, 0.18632491, 0.10689223])
message_box(
    f'Converting to "CIE XYZ" tristimulus values from given "IgPgTg" '
    f"colourspace values:\n\n\t{IgPgTg}"
)
print(colour.IgPgTg_to_XYZ(IgPgTg))

print("\n")

XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
message_box(
    f'Converting to "IPT" colourspace from given "CIE XYZ" tristimulus '
    f"values:\n\n\t{XYZ}"
)
print(colour.XYZ_to_IPT(XYZ))

print("\n")

IPT = np.array([0.36571124, -0.11114798, 0.01594746])
message_box(
    f'Converting to "CIE XYZ" tristimulus values from given "IPT" colourspace '
    f"values:\n\n\t{IPT}"
)
print(colour.IPT_to_XYZ(IPT))

print("\n")

message_box(
    f'Converting to "hdr-CIELab" colourspace from given "CIE XYZ" tristimulus '
    f"values:\n\n\t{XYZ}"
)
print(colour.XYZ_to_hdr_CIELab(XYZ))

print("\n")

Lab_hdr = np.array([48.26598942, -26.97517728, 4.99243377])
message_box(
    f'Converting to "CIE XYZ" tristimulus values from given "hdr-CIELab" '
    f"colourspace values:\n\n\t{Lab_hdr}"
)
print(colour.hdr_CIELab_to_XYZ(Lab_hdr))

print("\n")

message_box(
    f'Converting to "hdr-IPT" colourspace from given "CIE XYZ" tristimulus '
    f"values:\n\n\t{XYZ}"
)
print(colour.XYZ_to_hdr_IPT(XYZ))

print("\n")

IPT_hdr = np.array([46.4993815, -12.82251566, 1.85029518])
message_box(
    f'Converting to "CIE XYZ" tristimulus values from given "hdr-IPT" '
    f"colourspace values:\n\n\t{IPT_hdr}"
)
print(colour.hdr_IPT_to_XYZ(IPT_hdr))

print("\n")

message_box(
    f'Converting to "Jzazbz" colourspace from given "CIE XYZ" tristimulus '
    f"values:\n\n\t{XYZ}"
)
print(colour.XYZ_to_Jzazbz(XYZ))

print("\n")

Jzazbz = np.array([0.00357804, -0.00295507, 0.00038998])
message_box(
    f'Converting to "CIE XYZ" tristimulus values from given "Jzazbz" '
    f"colourspace values:\n\n\t{Jzazbz}"
)
print(colour.Jzazbz_to_XYZ(Jzazbz))

print("\n")

message_box(
    f'Converting to "OSA UCS" colourspace from given "CIE XYZ" tristimulus '
    f'values under the "CIE 1964 10 Degree Standard Observer":\n\n'
    f"\t{XYZ * 100}"
)
print(colour.XYZ_to_OSA_UCS(XYZ * 100))

print("\n")

Ljg = np.array([-4.4900683, 0.70305936, 3.03463664])
message_box(
    f'Converting to "CIE XYZ" tristimulus values under the '
    f'"CIE 1964 10 Degree Standard Observer" from "OSA UCS" colourspace:\n\n'
    f"\t{Ljg}"
)
print(colour.OSA_UCS_to_XYZ(Ljg))

print("\n")

message_box(
    f'Converting to "Oklab" colourspace from given "CIE XYZ" tristimulus '
    f"values:\n\n\t{XYZ}"
)
print(colour.XYZ_to_Oklab(XYZ))

print("\n")

Lab = np.array([0.51634019, 0.15469500, 0.06289579])
message_box(
    f'Converting to "CIE XYZ" tristimulus values from "Oklab" colourspace:\n\n'
    f"\t{Lab}"
)
print(colour.Oklab_to_XYZ(Lab))

print("\n")

XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
message_box(
    'Converting to the "ProLab" colourspace from given "CIE XYZ" '
    "tristimulus values:\n\n"
    f"\t{XYZ}"
)
print(colour.XYZ_to_ProLab(XYZ))

print("\n")

ProLab = np.array([0.36571124, -0.11114798, 0.01594746])
message_box(
    'Converting to "CIE XYZ" tristimulus values from given "ProLab" '
    "colourspace values:\n\n"
    f"\t{ProLab}"
)
print(colour.ProLab_to_XYZ(ProLab))

print("\n")

XYZ = np.array([19.01, 20.00, 21.78])
XYZ_w = np.array([95.05, 100.00, 108.88])
L_A = 318.31
Y_b = 20.0
surround = colour.VIEWING_CONDITIONS_CIECAM02["Average"]
specification = colour.XYZ_to_CIECAM02(XYZ, XYZ_w, L_A, Y_b, surround)
JMh = (specification.J, specification.M, specification.h)
message_box(
    f'Converting to "CAM02-UCS" colourspace from given "CIECAM02" colour '
    f'appearance model "JMh" correlates:\n\n\t{JMh}'
)
print(colour.JMh_CIECAM02_to_CAM02UCS(JMh))

print("\n")

message_box(
    f'Converting to "CAM02-UCS" colourspace from given "CIE XYZ" tristimulus '
    f"values:\n\n\t{JMh}"
)
print(colour.XYZ_to_CAM02UCS(XYZ / 100, XYZ_w=XYZ_w / 100, L_A=L_A, Y_b=Y_b))

print("\n")

specification = colour.XYZ_to_CAM16(XYZ, XYZ_w, L_A, Y_b, surround)
JMh = (specification.J, specification.M, specification.h)
message_box(
    f'Converting to "CAM16-UCS" colourspace from given "CAM16" colour '
    f'appearance model "JMh" correlates:\n\n\t{JMh}'
)
print(colour.JMh_CAM16_to_CAM16UCS(JMh))

message_box(
    f'Converting to "CAM16-UCS" colourspace from given "CIE XYZ" tristimulus '
    f"values:\n\n\t{JMh}"
)
print(colour.XYZ_to_CAM16UCS(XYZ / 100, XYZ_w=XYZ_w / 100, L_A=L_A, Y_b=Y_b))



"""Defines the unit tests for the :mod:`colour.plotting.blindness` module."""

import numpy as np
import unittest
from matplotlib.pyplot import Axes, Figure

from colour.plotting import plot_cvd_simulation_Machado2009

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestPlotCvdSimulationMachado2009",
]


class TestPlotCvdSimulationMachado2009(unittest.TestCase):
    """
    Define :func:`colour.plotting.blindness.plot_cvd_simulation_Machado2009`
    definition unit tests methods.
    """

    def test_plot_cvd_simulation_Machado2009(self):
        """
        Test :func:`colour.plotting.blindness.plot_cvd_simulation_Machado2009`
        definition.
        """

        figure, axes = plot_cvd_simulation_Machado2009(
            np.random.rand(32, 32, 3)
        )

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


if __name__ == "__main__":
    unittest.main()



"""
Common Plotting
===============

Defines the common plotting objects:

-   :func:`colour.plotting.colour_style`
-   :func:`colour.plotting.override_style`
-   :func:`colour.plotting.colour_cycle`
-   :func:`colour.plotting.artist`
-   :func:`colour.plotting.camera`
-   :func:`colour.plotting.decorate`
-   :func:`colour.plotting.boundaries`
-   :func:`colour.plotting.display`
-   :func:`colour.plotting.render`
-   :func:`colour.plotting.label_rectangles`
-   :func:`colour.plotting.uniform_axes3d`
-   :func:`colour.plotting.plot_single_colour_swatch`
-   :func:`colour.plotting.plot_multi_colour_swatches`
-   :func:`colour.plotting.plot_single_function`
-   :func:`colour.plotting.plot_multi_functions`
-   :func:`colour.plotting.plot_image`
"""

from __future__ import annotations

import functools
import itertools
import matplotlib
import matplotlib.cm
import matplotlib.patches as Patch
import matplotlib.pyplot as plt
import matplotlib.ticker
import numpy as np
import re
from dataclasses import dataclass, field
from functools import partial
from matplotlib.colors import LinearSegmentedColormap

from colour.characterisation import CCS_COLOURCHECKERS, ColourChecker
from colour.colorimetry import (
    MultiSpectralDistributions,
    MSDS_CMFS,
    SDS_ILLUMINANTS,
    SDS_LIGHT_SOURCES,
    SpectralDistribution,
)
from colour.hints import (
    Any,
    ArrayLike,
    Boolean,
    Callable,
    Dict,
    Floating,
    Integer,
    List,
    Literal,
    Mapping,
    NDArray,
    Optional,
    RegexFlag,
    Sequence,
    Tuple,
    TypedDict,
    Union,
    cast,
)
from colour.models import RGB_COLOURSPACES, RGB_Colourspace, XYZ_to_RGB
from colour.utilities import (
    CaseInsensitiveMapping,
    Structure,
    as_float_array,
    attest,
    first_item,
    is_sibling,
    is_string,
    filter_mapping,
    optional,
    runtime_warning,
    validate_method,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "CONSTANTS_COLOUR_STYLE",
    "CONSTANTS_ARROW_STYLE",
    "colour_style",
    "override_style",
    "XYZ_to_plotting_colourspace",
    "ColourSwatch",
    "colour_cycle",
    "KwargsArtist",
    "artist",
    "KwargsCamera",
    "camera",
    "KwargsRender",
    "render",
    "label_rectangles",
    "uniform_axes3d",
    "filter_passthrough",
    "filter_RGB_colourspaces",
    "filter_cmfs",
    "filter_illuminants",
    "filter_colour_checkers",
    "update_settings_collection",
    "plot_single_colour_swatch",
    "plot_multi_colour_swatches",
    "plot_single_function",
    "plot_multi_functions",
    "plot_image",
]

CONSTANTS_COLOUR_STYLE: Structure = Structure(
    **{
        "colour": Structure(
            **{
                "darkest": "#111111",
                "darker": "#222222",
                "dark": "#333333",
                "dim": "#505050",
                "average": "#808080",
                "light": "#D5D5D5",
                "bright": "#EEEEEE",
                "brighter": "#F0F0F0",
                "brightest": "#F5F5F5",
                "cycle": (
                    "#F44336",
                    "#9C27B0",
                    "#3F51B5",
                    "#03A9F4",
                    "#009688",
                    "#8BC34A",
                    "#FFEB3B",
                    "#FF9800",
                    "#795548",
                    "#607D8B",
                ),
                "map": LinearSegmentedColormap.from_list(
                    "colour",
                    (
                        "#F44336",
                        "#9C27B0",
                        "#3F51B5",
                        "#03A9F4",
                        "#009688",
                        "#8BC34A",
                        "#FFEB3B",
                        "#FF9800",
                        "#795548",
                        "#607D8B",
                    ),
                ),
                "colourspace": RGB_COLOURSPACES["sRGB"],
            }
        ),
        "opacity": Structure(**{"high": 0.75, "medium": 0.5, "low": 0.25}),
        "geometry": Structure(**{"long": 5, "medium": 2.5, "short": 1}),
        "hatch": Structure(
            **{
                "patterns": (
                    "\\\\",
                    "o",
                    "x",
                    ".",
                    "*",
                    "//",
                )
            }
        ),
        "zorder": Structure(
            {
                "background_polygon": -140,
                "background_scatter": -130,
                "background_line": -120,
                "background_annotation": -110,
                "background_label": -100,
                "midground_polygon": -90,
                "midground_scatter": -80,
                "midground_line": -70,
                "midground_annotation": -60,
                "midground_label": -50,
                "foreground_polygon": -40,
                "foreground_scatter": -30,
                "foreground_line": -20,
                "foreground_annotation": -10,
                "foreground_label": 0,
            }
        ),
    }
)
"""Various defaults settings used across the plotting sub-package."""

CONSTANTS_ARROW_STYLE: Structure = Structure(
    **{
        "color": CONSTANTS_COLOUR_STYLE.colour.dark,
        "headwidth": CONSTANTS_COLOUR_STYLE.geometry.short * 4,
        "headlength": CONSTANTS_COLOUR_STYLE.geometry.long,
        "width": CONSTANTS_COLOUR_STYLE.geometry.short * 0.5,
        "shrink": CONSTANTS_COLOUR_STYLE.geometry.short * 0.1,
        "connectionstyle": "arc3,rad=-0.2",
    }
)
"""Annotation arrow settings used across the plotting sub-package."""


def colour_style(use_style: Boolean = True) -> Dict:
    """
    Return *Colour* plotting style.

    Parameters
    ----------
    use_style
        Whether to use the style and load it into *Matplotlib*.

    Returns
    -------
    :class:`dict`
        *Colour* style.
    """

    constants = CONSTANTS_COLOUR_STYLE
    style = {
        # Figure Size Settings
        "figure.figsize": (12.80, 7.20),
        "figure.dpi": 100,
        "savefig.dpi": 100,
        "savefig.bbox": "standard",
        # Font Settings
        # 'font.size': 12,
        "axes.titlesize": "x-large",
        "axes.labelsize": "larger",
        "legend.fontsize": "small",
        "xtick.labelsize": "medium",
        "ytick.labelsize": "medium",
        # Text Settings
        "text.color": constants.colour.darkest,
        # Tick Settings
        "xtick.top": False,
        "xtick.bottom": True,
        "ytick.right": False,
        "ytick.left": True,
        "xtick.minor.visible": True,
        "ytick.minor.visible": True,
        "xtick.direction": "out",
        "ytick.direction": "out",
        "xtick.major.size": constants.geometry.long * 1.25,
        "xtick.minor.size": constants.geometry.long * 0.75,
        "ytick.major.size": constants.geometry.long * 1.25,
        "ytick.minor.size": constants.geometry.long * 0.75,
        "xtick.major.width": constants.geometry.short,
        "xtick.minor.width": constants.geometry.short,
        "ytick.major.width": constants.geometry.short,
        "ytick.minor.width": constants.geometry.short,
        # Spine Settings
        "axes.linewidth": constants.geometry.short,
        "axes.edgecolor": constants.colour.dark,
        # Title Settings
        "axes.titlepad": plt.rcParams["font.size"] * 0.75,
        # Axes Settings
        "axes.facecolor": constants.colour.brightest,
        "axes.grid": True,
        "axes.grid.which": "major",
        "axes.grid.axis": "both",
        # Grid Settings
        "axes.axisbelow": True,
        "grid.linewidth": constants.geometry.short * 0.5,
        "grid.linestyle": "--",
        "grid.color": constants.colour.light,
        # Legend
        "legend.frameon": True,
        "legend.framealpha": constants.opacity.high,
        "legend.fancybox": False,
        "legend.facecolor": constants.colour.brighter,
        "legend.borderpad": constants.geometry.short * 0.5,
        # Lines
        "lines.linewidth": constants.geometry.short,
        "lines.markersize": constants.geometry.short * 3,
        "lines.markeredgewidth": constants.geometry.short * 0.75,
        # Cycle
        "axes.prop_cycle": matplotlib.cycler(color=constants.colour.cycle),
    }

    if use_style:
        plt.rcParams.update(style)

    return style


def override_style(**kwargs: Any) -> Callable:
    """
    Decorate a function to override *Matplotlib* style.

    Other Parameters
    ----------------
    kwargs
        Keywords arguments.

    Returns
    -------
    Callable

    Examples
    --------
    >>> @override_style(**{'text.color': 'red'})
    ... def f():
    ...     plt.text(0.5, 0.5, 'This is a text!')
    ...     plt.show()
    >>> f()  # doctest: +SKIP
    """

    keywords = dict(kwargs)

    def wrapper(function: Callable) -> Callable:
        """Wrap given function wrapper."""

        @functools.wraps(function)
        def wrapped(*args: Any, **kwargs: Any) -> Any:
            """Wrap given function."""

            keywords.update(kwargs)

            style_overrides = {
                key: value
                for key, value in keywords.items()
                if key in plt.rcParams.keys()
            }

            with plt.style.context(style_overrides):
                return function(*args, **kwargs)

        return wrapped

    return wrapper


def XYZ_to_plotting_colourspace(
    XYZ: ArrayLike,
    illuminant: ArrayLike = RGB_COLOURSPACES["sRGB"].whitepoint,
    chromatic_adaptation_transform: Union[
        Literal[
            "Bianco 2010",
            "Bianco PC 2010",
            "Bradford",
            "CAT02 Brill 2008",
            "CAT02",
            "CAT16",
            "CMCCAT2000",
            "CMCCAT97",
            "Fairchild",
            "Sharp",
            "Von Kries",
            "XYZ Scaling",
        ],
        str,
    ] = "CAT02",
    apply_cctf_encoding: Boolean = True,
) -> NDArray:
    """
    Convert from *CIE XYZ* tristimulus values to the default plotting
    colourspace.

    Parameters
    ----------
    XYZ
        *CIE XYZ* tristimulus values.
    illuminant
        Source illuminant chromaticity coordinates.
    chromatic_adaptation_transform
        *Chromatic adaptation* transform.
    apply_cctf_encoding
        Apply the default plotting colourspace encoding colour component
        transfer function / opto-electronic transfer function.

    Returns
    -------
    :class:`numpy.ndarray`
        Default plotting colourspace colour array.

    Examples
    --------
    >>> import numpy as np
    >>> XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
    >>> XYZ_to_plotting_colourspace(XYZ)  # doctest: +ELLIPSIS
    array([ 0.7057393...,  0.1924826...,  0.2235416...])
    """

    return XYZ_to_RGB(
        XYZ,
        illuminant,
        CONSTANTS_COLOUR_STYLE.colour.colourspace.whitepoint,
        CONSTANTS_COLOUR_STYLE.colour.colourspace.matrix_XYZ_to_RGB,
        chromatic_adaptation_transform,
        CONSTANTS_COLOUR_STYLE.colour.colourspace.cctf_encoding
        if apply_cctf_encoding
        else None,
    )


@dataclass
class ColourSwatch:
    """
    Define a data structure for a colour swatch.

    Parameters
    ----------
    RGB
        RGB Colour.
    name
        Colour name.
    """

    RGB: ArrayLike
    name: Optional[str] = field(default_factory=lambda: None)


def colour_cycle(**kwargs: Any) -> itertools.cycle:
    """
    Return a colour cycle iterator using given colour map.

    Other Parameters
    ----------------
    colour_cycle_map
        Matplotlib colourmap name.
    colour_cycle_count
        Colours count to pick in the colourmap.

    Returns
    -------
    :class:`itertools.cycle`
        Colour cycle iterator.
    """

    settings = Structure(
        **{
            "colour_cycle_map": CONSTANTS_COLOUR_STYLE.colour.map,
            "colour_cycle_count": len(CONSTANTS_COLOUR_STYLE.colour.cycle),
        }
    )
    settings.update(kwargs)

    samples = np.linspace(0, 1, settings.colour_cycle_count)
    if isinstance(settings.colour_cycle_map, LinearSegmentedColormap):
        cycle = settings.colour_cycle_map(samples)
    else:
        cycle = getattr(plt.cm, settings.colour_cycle_map)(samples)

    return itertools.cycle(cycle)


class KwargsArtist(TypedDict):
    """
    Define the keyword argument types for the :func:`colour.plotting.artist`
    definition.

    Parameters
    ----------
    axes
        Axes that will be passed through without creating a new figure.
    uniform
        Whether to create the figure with an equal aspect ratio.
    """

    axes: plt.Axes
    uniform: Boolean


def artist(**kwargs: Union[KwargsArtist, Any]) -> Tuple[plt.Figure, plt.Axes]:
    """
    Return the current figure and its axes or creates a new one.

    Other Parameters
    ----------------
    kwargs
        {:func:`colour.plotting.common.KwargsArtist`},
        See the documentation of the previously listed class.

    Returns
    -------
    :class:`tuple`
        Current figure and axes.
    """

    width, height = plt.rcParams["figure.figsize"]

    figure_size = (width, width) if kwargs.get("uniform") else (width, height)

    axes = kwargs.get("axes")
    if axes is None:
        figure = plt.figure(figsize=figure_size)

        return figure, figure.gca()
    else:
        return plt.gcf(), axes


class KwargsCamera(TypedDict):
    """
    Define the keyword argument types for the :func:`colour.plotting.camera`
    definition.

    Parameters
    ----------
    figure
        Figure to apply the render elements onto.
    axes
        Axes to apply the render elements onto.
    azimuth
        Camera azimuth.
    elevation
        Camera elevation.
    camera_aspect
        Matplotlib axes aspect. Default is *equal*.
    """

    figure: plt.Figure
    axes: plt.Axes
    azimuth: Optional[Floating]
    elevation: Optional[Floating]
    camera_aspect: Union[Literal["equal"], str]


def camera(**kwargs: Union[KwargsCamera, Any]) -> Tuple[plt.Figure, plt.Axes]:
    """
    Set the camera settings.

    Other Parameters
    ----------------
    kwargs
        {:func:`colour.plotting.common.KwargsCamera`},
        See the documentation of the previously listed class.

    Returns
    -------
    :class:`tuple`
        Current figure and axes.
    """

    figure = cast(plt.Figure, kwargs.get("figure", plt.gcf()))
    axes = cast(plt.Axes, kwargs.get("axes", plt.gca()))

    settings = Structure(
        **{"camera_aspect": "equal", "elevation": None, "azimuth": None}
    )
    settings.update(kwargs)

    if settings.camera_aspect == "equal":
        uniform_axes3d(axes=axes)

    axes.view_init(elev=settings.elevation, azim=settings.azimuth)

    return figure, axes


class KwargsRender(TypedDict):
    """
    Define the keyword argument types for the :func:`colour.plotting.render`
    definition.

    Parameters
    ----------
    figure
        Figure to apply the render elements onto.
    axes
        Axes to apply the render elements onto.
    filename
        Figure will be saved using given ``filename`` argument.
    standalone
        Whether to show the figure and call :func:`matplotlib.pyplot.show`
        definition.
    aspect
        Matplotlib axes aspect.
    axes_visible
        Whether the axes are visible. Default is *True*.
    bounding_box
        Array defining current axes limits such
        `bounding_box = (x min, x max, y min, y max)`.
    tight_layout
        Whether to invoke the :func:`matplotlib.pyplot.tight_layout`
        definition.
    legend
        Whether to display the legend. Default is *False*.
    legend_columns
        Number of columns in the legend. Default is *1*.
    transparent_background
        Whether to turn off the background patch. Default is *True*.
    title
        Figure title.
    wrap_title
        Whether to wrap the figure title. Default is *True*.
    x_label
        *X* axis label.
    y_label
        *Y* axis label.
    x_ticker
        Whether to display the *X* axis ticker. Default is *True*.
    y_ticker
        Whether to display the *Y* axis ticker. Default is *True*.
    """

    figure: plt.Figure
    axes: plt.Axes
    filename: str
    standalone: Boolean
    aspect: Union[Literal["auto", "equal"], Floating]
    axes_visible: Boolean
    bounding_box: ArrayLike
    tight_layout: Boolean
    legend: Boolean
    legend_columns: Integer
    transparent_background: Boolean
    title: str
    wrap_title: Boolean
    x_label: str
    y_label: str
    x_ticker: Boolean
    y_ticker: Boolean


def render(**kwargs: Union[KwargsRender, Any]) -> Tuple[plt.Figure, plt.Axes]:
    """
    Render the current figure while adjusting various settings such as the
    bounding box, the title or background transparency.

    Other Parameters
    ----------------
    kwargs
        {:func:`colour.plotting.common.KwargsRender`},
        See the documentation of the previously listed class.

    Returns
    -------
    :class:`tuple`
        Current figure and axes.
    """

    figure = cast(plt.Figure, kwargs.get("figure", plt.gcf()))
    axes = cast(plt.Axes, kwargs.get("axes", plt.gca()))

    settings = Structure(
        **{
            "filename": None,
            "standalone": True,
            "aspect": None,
            "axes_visible": True,
            "bounding_box": None,
            "tight_layout": True,
            "legend": False,
            "legend_columns": 1,
            "transparent_background": True,
            "title": None,
            "wrap_title": True,
            "x_label": None,
            "y_label": None,
            "x_ticker": True,
            "y_ticker": True,
        }
    )
    settings.update(kwargs)

    if settings.aspect:
        axes.set_aspect(settings.aspect)
    if not settings.axes_visible:
        axes.set_axis_off()
    if settings.bounding_box:
        axes.set_xlim(settings.bounding_box[0], settings.bounding_box[1])
        axes.set_ylim(settings.bounding_box[2], settings.bounding_box[3])

    if settings.title:
        axes.set_title(settings.title, wrap=settings.wrap_title)
    if settings.x_label:
        axes.set_xlabel(settings.x_label)
    if settings.y_label:
        axes.set_ylabel(settings.y_label)
    if not settings.x_ticker:
        axes.set_xticks([])
    if not settings.y_ticker:
        axes.set_yticks([])
    if settings.legend:
        axes.legend(ncol=settings.legend_columns)

    if settings.tight_layout:
        figure.tight_layout()

    if settings.transparent_background:
        figure.patch.set_alpha(0)
    if settings.standalone:
        if settings.filename is not None:
            figure.savefig(settings.filename)
        else:
            plt.show()

    return figure, axes


def label_rectangles(
    labels: Sequence[str],
    rectangles: Sequence[Patch],
    rotation: Union[Literal["horizontal", "vertical"], str] = "vertical",
    text_size: Floating = 10,
    offset: Optional[ArrayLike] = None,
    **kwargs: Any,
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Add labels above given rectangles.

    Parameters
    ----------
    labels
        Labels to display.
    rectangles
        Rectangles to used to set the labels value and position.
    rotation
        Labels orientation.
    text_size
        Labels text size.
    offset
        Labels offset as percentages of the largest rectangle dimensions.

    Other Parameters
    ----------------
    figure
        Figure to apply the render elements onto.
    axes
        Axes to apply the render elements onto.

    Returns
    -------
    :class:`tuple`
        Current figure and axes.
    """

    rotation = validate_method(
        rotation,
        ["horizontal", "vertical"],
        '"{0}" rotation is invalid, it must be one of {1}!',
    )

    figure = kwargs.get("figure", plt.gcf())
    axes = kwargs.get("axes", plt.gca())

    offset = as_float_array(cast(ArrayLike, optional(offset, (0.0, 0.025))))

    x_m, y_m = 0, 0
    for rectangle in rectangles:
        x_m = max(x_m, rectangle.get_width())
        y_m = max(y_m, rectangle.get_height())

    for i, rectangle in enumerate(rectangles):
        x = rectangle.get_x()
        height = rectangle.get_height()
        width = rectangle.get_width()
        ha = "center"
        va = "bottom"
        axes.text(
            x + width / 2 + offset[0] * width,
            height + offset[1] * y_m,
            labels[i],
            ha=ha,
            va=va,
            rotation=rotation,
            fontsize=text_size,
            clip_on=True,
            zorder=CONSTANTS_COLOUR_STYLE.zorder.midground_label,
        )

    return figure, axes


def uniform_axes3d(**kwargs: Any) -> Tuple[plt.Figure, plt.Axes]:
    """
    Set equal aspect ratio to given 3d axes.

    Other Parameters
    ----------------
    figure
        Figure to apply the render elements onto.
    axes
        Axes to apply the render elements onto.

    Returns
    -------
    :class:`tuple`
        Current figure and axes.
    """

    figure = kwargs.get("figure", plt.gcf())
    axes = kwargs.get("axes", plt.gca())

    try:  # pragma: no cover
        # TODO: Reassess according to
        # https://github.com/matplotlib/matplotlib/issues/1077
        axes.set_aspect("equal")
    except NotImplementedError:  # pragma: no cover
        pass

    extents = np.array([getattr(axes, f"get_{axis}lim")() for axis in "xyz"])

    centers = np.mean(extents, axis=1)
    extent = np.max(np.abs(extents[..., 1] - extents[..., 0]))

    for center, axis in zip(centers, "xyz"):
        getattr(axes, f"set_{axis}lim")(
            center - extent / 2, center + extent / 2
        )

    return figure, axes


def filter_passthrough(
    mapping: Mapping,
    filterers: Union[Any, str, Sequence[Union[Any, str]]],
    anchors: Boolean = True,
    allow_non_siblings: Boolean = True,
    flags: Union[Integer, RegexFlag] = re.IGNORECASE,
) -> Dict:
    """
    Return mapping objects matching given filterers while passing through
    class instances whose type is one of the mapping element types.

    This definition allows passing custom but compatible objects to the various
    plotting definitions that by default expect the key from a dataset element.

    For example, a typical call to :func:`colour.plotting.\
plot_multi_illuminant_sds` definition with a regex pattern automatically
    anchored at boundaries by default is as follows:

    >>> import colour
    >>> colour.plotting.plot_multi_illuminant_sds(['A'])
    ... # doctest: +SKIP

    Here, `'A'` is by default anchored at boundaries and transformed into
    `'^A$'`. Note that because it is a regex pattern, special characters such
    as parenthesis must be escaped: `'Adobe RGB (1998)'` must be written
    `'Adobe RGB \\(1998\\)'` instead.

    With the previous example, t is also possible to pass a custom spectral
    distribution as follows:

    >>> data = {
    ...     500: 0.0651,
    ...     520: 0.0705,
    ...     540: 0.0772,
    ...     560: 0.0870,
    ...     580: 0.1128,
    ...     600: 0.1360
    ... }
    >>> colour.plotting.plot_multi_illuminant_sds(
    ...     ['A', colour.SpectralDistribution(data)])
    ... # doctest: +SKIP

    Similarly, a typical call to :func:`colour.plotting.\
plot_planckian_locus_in_chromaticity_diagram_CIE1931` definition is as follows:

    >>> colour.plotting.plot_planckian_locus_in_chromaticity_diagram_CIE1931(
    ...     ['A'])
    ... # doctest: +SKIP

    But it is also possible to pass a custom whitepoint as follows:

    >>> colour.plotting.plot_planckian_locus_in_chromaticity_diagram_CIE1931(
    ...     ['A', {'Custom': np.array([1 / 3 + 0.05, 1 / 3 + 0.05])}])
    ... # doctest: +SKIP

    Parameters
    ----------
    mapping
        Mapping to filter.
    filterers
        Filterer or object class instance (which is passed through directly if
        its type is one of the mapping element types) or list
        of filterers.
    anchors
        Whether to use Regex line anchors, i.e. *^* and *$* are added,
        surrounding the filterers patterns.
    allow_non_siblings
        Whether to allow non-siblings to be also passed through.
    flags
        Regex flags.

    Returns
    -------
    :class:`dict`
        Filtered mapping.
    """

    if is_string(filterers):
        filterers = [filterers]
    elif not isinstance(filterers, (list, tuple)):
        filterers = [filterers]

    string_filterers: List[str] = [
        cast(str, filterer) for filterer in filterers if is_string(filterer)
    ]

    object_filterers: List[Any] = [
        filterer for filterer in filterers if is_sibling(filterer, mapping)
    ]

    if allow_non_siblings:
        non_siblings = [
            filterer
            for filterer in filterers
            if filterer not in string_filterers
            and filterer not in object_filterers
        ]

        if non_siblings:
            runtime_warning(
                f'Non-sibling elements are passed-through: "{non_siblings}"'
            )

            object_filterers.extend(non_siblings)

    filtered_mapping = filter_mapping(
        mapping, string_filterers, anchors, flags
    )

    for filterer in object_filterers:
        # TODO: Consider using "MutableMapping" here.
        if isinstance(filterer, (dict, CaseInsensitiveMapping)):
            for key, value in filterer.items():
                filtered_mapping[key] = value
        else:
            try:
                name = filterer.name
            except AttributeError:
                try:
                    name = filterer.__name__
                except AttributeError:
                    name = str(id(filterer))

            filtered_mapping[name] = filterer

    return filtered_mapping


def filter_RGB_colourspaces(
    filterers: Union[
        RGB_Colourspace, str, Sequence[Union[RGB_Colourspace, str]]
    ],
    anchors: Boolean = True,
    allow_non_siblings: Boolean = True,
    flags: Union[Integer, RegexFlag] = re.IGNORECASE,
) -> Dict[str, RGB_Colourspace]:
    """
    Return the *RGB* colourspaces matching given filterers.

    Parameters
    ----------
    filterers
        Filterer or :class:`colour.RGB_Colourspace` class instance (which is
        passed through directly if its type is one of the mapping element
        types) or list of filterers. ``filterers`` elements can also be of any
        form supported by the :func:`colour.plotting.filter_passthrough`
        definition.
    anchors
        Whether to use Regex line anchors, i.e. *^* and *$* are added,
        surrounding the filterers patterns.
    allow_non_siblings
        Whether to allow non-siblings to be also passed through.
    flags
        Regex flags.

    Returns
    -------
    :class:`dict`
        Filtered *RGB* colourspaces.
    """

    return filter_passthrough(
        RGB_COLOURSPACES, filterers, anchors, allow_non_siblings, flags
    )


def filter_cmfs(
    filterers: Union[
        MultiSpectralDistributions,
        str,
        Sequence[Union[MultiSpectralDistributions, str]],
    ],
    anchors: Boolean = True,
    allow_non_siblings: Boolean = True,
    flags: Union[Integer, RegexFlag] = re.IGNORECASE,
) -> Dict[str, MultiSpectralDistributions]:
    """
    Return the colour matching functions matching given filterers.

    Parameters
    ----------
    filterers
        Filterer or :class:`colour.LMS_ConeFundamentals`,
        :class:`colour.RGB_ColourMatchingFunctions` or
        :class:`colour.XYZ_ColourMatchingFunctions` class instance (which is
        passed through directly if its type is one of the mapping element
        types) or list of filterers. ``filterers`` elements can also be of any
        form supported by the :func:`colour.plotting.filter_passthrough`
        definition.
    anchors
        Whether to use Regex line anchors, i.e. *^* and *$* are added,
        surrounding the filterers patterns.
    allow_non_siblings
        Whether to allow non-siblings to be also passed through.
    flags
        Regex flags.

    Returns
    -------
    :class:`dict`
        Filtered colour matching functions.
    """

    return filter_passthrough(
        MSDS_CMFS, filterers, anchors, allow_non_siblings, flags
    )


def filter_illuminants(
    filterers: Union[
        SpectralDistribution, str, Sequence[Union[SpectralDistribution, str]]
    ],
    anchors: Boolean = True,
    allow_non_siblings: Boolean = True,
    flags: Union[Integer, RegexFlag] = re.IGNORECASE,
) -> Dict[str, SpectralDistribution]:
    """
    Return the illuminants matching given filterers.

    Parameters
    ----------
    filterers
        Filterer or :class:`colour.SpectralDistribution` class instance
        (which is passed through directly if its type is one of the mapping
        element types) or list of filterers. ``filterers`` elements can also be
        of any form supported by the :func:`colour.plotting.filter_passthrough`
        definition.
    anchors
        Whether to use Regex line anchors, i.e. *^* and *$* are added,
        surrounding the filterers patterns.
    allow_non_siblings
        Whether to allow non-siblings to be also passed through.
    flags
        Regex flags.

    Returns
    -------
    :class:`dict`
        Filtered illuminants.
    """

    illuminants = {}

    illuminants.update(
        filter_passthrough(
            SDS_ILLUMINANTS, filterers, anchors, allow_non_siblings, flags
        )
    )

    illuminants.update(
        filter_passthrough(
            SDS_LIGHT_SOURCES, filterers, anchors, allow_non_siblings, flags
        )
    )

    return illuminants


def filter_colour_checkers(
    filterers: Union[ColourChecker, str, Sequence[Union[ColourChecker, str]]],
    anchors: Boolean = True,
    allow_non_siblings: Boolean = True,
    flags: Union[Integer, RegexFlag] = re.IGNORECASE,
) -> Dict[str, ColourChecker]:
    """
    Return the colour checkers matching given filterers.

    Parameters
    ----------
    filterers
        Filterer or :class:`colour.characterisation.ColourChecker` class
        instance (which is passed through directly if its type is one of the
        mapping element types) or list of filterers. ``filterers`` elements
        can also be of any form supported by the
        :func:`colour.plotting.filter_passthrough` definition.
    anchors
        Whether to use Regex line anchors, i.e. *^* and *$* are added,
        surrounding the filterers patterns.
    allow_non_siblings
        Whether to allow non-siblings to be also passed through.
    flags
        Regex flags.

    Returns
    -------
    :class:`dict`
        Filtered colour checkers.
    """

    return filter_passthrough(
        CCS_COLOURCHECKERS, filterers, anchors, allow_non_siblings, flags
    )


def update_settings_collection(
    settings_collection: Union[Dict, List[Dict]],
    keyword_arguments: Union[Dict, List[Dict]],
    expected_count: Integer,
):
    """
    Update given settings collection, *in-place*, with given keyword arguments
    and expected count of settings collection elements.

    Parameters
    ----------
    settings_collection
        Settings collection to update.
    keyword_arguments
        Keyword arguments to update the settings collection.
    expected_count
        Expected count of settings collection elements.

    Examples
    --------
    >>> settings_collection = [{1: 2}, {3: 4}]
    >>> keyword_arguments = {5 : 6}
    >>> update_settings_collection(settings_collection, keyword_arguments, 2)
    >>> print(settings_collection)
    [{1: 2, 5: 6}, {3: 4, 5: 6}]
    >>> settings_collection = [{1: 2}, {3: 4}]
    >>> keyword_arguments = [{5 : 6}, {7: 8}]
    >>> update_settings_collection(settings_collection, keyword_arguments, 2)
    >>> print(settings_collection)
    [{1: 2, 5: 6}, {3: 4, 7: 8}]
    """

    if not isinstance(keyword_arguments, dict):
        attest(
            len(keyword_arguments) == expected_count,
            "Multiple keyword arguments defined, but they do not "
            "match the expected count!",
        )

    for i, settings in enumerate(settings_collection):
        if isinstance(keyword_arguments, dict):
            settings.update(keyword_arguments)
        else:
            settings.update(keyword_arguments[i])


@override_style(
    **{
        "axes.grid": False,
        "xtick.bottom": False,
        "ytick.left": False,
        "xtick.labelbottom": False,
        "ytick.labelleft": False,
    }
)
def plot_single_colour_swatch(
    colour_swatch: Union[ArrayLike, ColourSwatch], **kwargs: Any
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Plot given colour swatch.

    Parameters
    ----------
    colour_swatch
        Colour swatch, either a regular `ArrayLike` or a
        :class:`colour.plotting.ColourSwatch` class instance.

    Other Parameters
    ----------------
    kwargs
        {:func:`colour.plotting.artist`,
        :func:`colour.plotting.plot_multi_colour_swatches`,
        :func:`colour.plotting.render`},
        See the documentation of the previously listed definitions.

    Returns
    -------
    :class:`tuple`
        Current figure and axes.

    Examples
    --------
    >>> RGB = ColourSwatch((0.45620519, 0.03081071, 0.04091952))
    >>> plot_single_colour_swatch(RGB)  # doctest: +ELLIPSIS
    (<Figure size ... with 1 Axes>, <...AxesSubplot...>)

    .. image:: ../_static/Plotting_Plot_Single_Colour_Swatch.png
        :align: center
        :alt: plot_single_colour_swatch
    """

    return plot_multi_colour_swatches((colour_swatch,), **kwargs)


@override_style(
    **{
        "axes.grid": False,
        "xtick.bottom": False,
        "ytick.left": False,
        "xtick.labelbottom": False,
        "ytick.labelleft": False,
    }
)
def plot_multi_colour_swatches(
    colour_swatches: Sequence[Union[ArrayLike, ColourSwatch]],
    width: Floating = 1,
    height: Floating = 1,
    spacing: Floating = 0,
    columns: Optional[Integer] = None,
    direction: Union[Literal["+y", "-y"], str] = "+y",
    text_kwargs: Optional[Dict] = None,
    background_colour: ArrayLike = (1.0, 1.0, 1.0),
    compare_swatches: Optional[
        Union[Literal["Diagonal", "Stacked"], str]
    ] = None,
    **kwargs: Any,
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Plot given colours swatches.

    Parameters
    ----------
    colour_swatches
        Colour swatch sequence, either a regular `ArrayLike` or a sequence of
        :class:`colour.plotting.ColourSwatch` class instances.
    width
        Colour swatch width.
    height
        Colour swatch height.
    spacing
        Colour swatches spacing.
    columns
        Colour swatches columns count, defaults to the colour swatch count or
        half of it if comparing.
    direction
        Row stacking direction.
    text_kwargs
        Keyword arguments for the :func:`matplotlib.pyplot.text` definition.
        The following special keywords can also be used:

        -   ``offset``: Sets the text offset.
        -   ``visible``: Sets the text visibility.
    background_colour
        Background colour.
    compare_swatches
        Whether to compare the swatches, in which case the colour swatch
        count must be an even number with alternating reference colour swatches
        and test colour swatches. *Stacked* will draw the test colour swatch in
        the center of the reference colour swatch, *Diagonal* will draw
        the reference colour swatch in the upper left diagonal area and the
        test colour swatch in the bottom right diagonal area.

    Other Parameters
    ----------------
    kwargs
        {:func:`colour.plotting.artist`, :func:`colour.plotting.render`},
        See the documentation of the previously listed definitions.

    Returns
    -------
    :class:`tuple`
        Current figure and axes.

    Examples
    --------
    >>> RGB_1 = ColourSwatch((0.45293517, 0.31732158, 0.26414773))
    >>> RGB_2 = ColourSwatch((0.77875824, 0.57726450, 0.50453169))
    >>> plot_multi_colour_swatches([RGB_1, RGB_2])  # doctest: +ELLIPSIS
    (<Figure size ... with 1 Axes>, <...AxesSubplot...>)

    .. image:: ../_static/Plotting_Plot_Multi_Colour_Swatches.png
        :align: center
        :alt: plot_multi_colour_swatches
    """

    direction = validate_method(
        direction,
        ["+y", "-y"],
        '"{0}" direction is invalid, it must be one of {1}!',
    )

    if compare_swatches is not None:
        compare_swatches = validate_method(
            compare_swatches,
            ["Diagonal", "Stacked"],
            '"{0}" compare swatches method is invalid, it must be one of {1}!',
        )

    _figure, axes = artist(**kwargs)

    # Handling case where `colour_swatches` is a regular *ArrayLike*.
    colour_swatches = list(colour_swatches)
    colour_swatches_converted = []
    if not isinstance(first_item(colour_swatches), ColourSwatch):
        for i, colour_swatch in enumerate(
            as_float_array(cast(ArrayLike, colour_swatches)).reshape([-1, 3])
        ):
            colour_swatches_converted.append(ColourSwatch(colour_swatch))
    else:
        colour_swatches_converted = cast(List[ColourSwatch], colour_swatches)

    colour_swatches = colour_swatches_converted

    if compare_swatches is not None:
        attest(
            len(colour_swatches) % 2 == 0,
            "Cannot compare an odd number of colour swatches!",
        )

        colour_swatches_reference = colour_swatches[0::2]
        colour_swatches_test = colour_swatches[1::2]
    else:
        colour_swatches_reference = colour_swatches_test = colour_swatches

    columns = optional(columns, len(colour_swatches_reference))

    text_settings = {
        "offset": 0.05,
        "visible": True,
        "zorder": CONSTANTS_COLOUR_STYLE.zorder.midground_label,
    }
    if text_kwargs is not None:
        text_settings.update(text_kwargs)
    text_offset = text_settings.pop("offset")

    offset_X: Floating = 0
    offset_Y: Floating = 0
    x_min, x_max, y_min, y_max = 0, width, 0, height
    y = 1 if direction == "+y" else -1
    for i, colour_swatch in enumerate(colour_swatches_reference):
        if i % columns == 0 and i != 0:
            offset_X = 0
            offset_Y += (height + spacing) * y

        x_0, x_1 = offset_X, offset_X + width
        y_0, y_1 = offset_Y, offset_Y + height * y

        axes.fill(
            (x_0, x_1, x_1, x_0),
            (y_0, y_0, y_1, y_1),
            color=np.clip(colour_swatches_reference[i].RGB, 0, 1),
            zorder=CONSTANTS_COLOUR_STYLE.zorder.midground_polygon,
        )

        if compare_swatches == "stacked":
            margin_X = width * 0.25
            margin_Y = height * 0.25
            axes.fill(
                (
                    x_0 + margin_X,
                    x_1 - margin_X,
                    x_1 - margin_X,
                    x_0 + margin_X,
                ),
                (
                    y_0 + margin_Y * y,
                    y_0 + margin_Y * y,
                    y_1 - margin_Y * y,
                    y_1 - margin_Y * y,
                ),
                color=np.clip(colour_swatches_test[i].RGB, 0, 1),
                zorder=CONSTANTS_COLOUR_STYLE.zorder.midground_polygon,
            )
        else:
            axes.fill(
                (x_0, x_1, x_1),
                (y_0, y_0, y_1),
                color=np.clip(colour_swatches_test[i].RGB, 0, 1),
                zorder=CONSTANTS_COLOUR_STYLE.zorder.midground_polygon,
            )

        if colour_swatch.name is not None and text_settings["visible"]:
            axes.text(
                x_0 + text_offset,
                y_0 + text_offset * y,
                colour_swatch.name,
                verticalalignment="bottom" if y == 1 else "top",
                clip_on=True,
                **text_settings,
            )

        offset_X += width + spacing

    x_max = min(len(colour_swatches), int(columns))
    x_max = x_max * width + x_max * spacing - spacing
    y_max = offset_Y

    axes.patch.set_facecolor(background_colour)

    if y == 1:
        bounding_box = [
            x_min - spacing,
            x_max + spacing,
            y_min - spacing,
            y_max + spacing + height,
        ]
    else:
        bounding_box = [
            x_min - spacing,
            x_max + spacing,
            y_max - spacing - height,
            y_min + spacing,
        ]

    settings: Dict[str, Any] = {
        "axes": axes,
        "bounding_box": bounding_box,
        "aspect": "equal",
    }
    settings.update(kwargs)

    return render(**settings)


@override_style()
def plot_single_function(
    function: Callable,
    samples: Optional[ArrayLike] = None,
    log_x: Optional[Integer] = None,
    log_y: Optional[Integer] = None,
    plot_kwargs: Optional[Union[Dict, List[Dict]]] = None,
    **kwargs: Any,
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Plot given function.

    Parameters
    ----------
    function
        Function to plot.
    samples
        Samples to evaluate the functions with.
    log_x
        Log base to use for the *x* axis scale, if *None*, the *x* axis scale
        will be linear.
    log_y
        Log base to use for the *y* axis scale, if *None*, the *y* axis scale
        will be linear.
    plot_kwargs
        Keyword arguments for the :func:`matplotlib.pyplot.plot` definition,
        used to control the style of the plotted function.

    Other Parameters
    ----------------
    kwargs
        {:func:`colour.plotting.artist`,
        :func:`colour.plotting.plot_multi_functions`,
        :func:`colour.plotting.render`},
        See the documentation of the previously listed definitions.

    Returns
    -------
    :class:`tuple`
        Current figure and axes.

    Examples
    --------
    >>> from colour.models import gamma_function
    >>> plot_single_function(partial(gamma_function, exponent=1 / 2.2))
    ... # doctest: +ELLIPSIS
    (<Figure size ... with 1 Axes>, <...AxesSubplot...>)

    .. image:: ../_static/Plotting_Plot_Single_Function.png
        :align: center
        :alt: plot_single_function
    """

    try:
        name = function.__name__
    except AttributeError:
        name = "Unnamed"

    settings: Dict[str, Any] = {
        "title": f"{name} - Function",
        "legend": False,
    }
    settings.update(kwargs)

    return plot_multi_functions(
        {name: function}, samples, log_x, log_y, plot_kwargs, **settings
    )


@override_style()
def plot_multi_functions(
    functions: Dict[str, Callable],
    samples: Optional[ArrayLike] = None,
    log_x: Optional[Integer] = None,
    log_y: Optional[Integer] = None,
    plot_kwargs: Optional[Union[Dict, List[Dict]]] = None,
    **kwargs: Any,
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Plot given functions.

    Parameters
    ----------
    functions
        Functions to plot.
    samples
        Samples to evaluate the functions with.
    log_x
        Log base to use for the *x* axis scale, if *None*, the *x* axis scale
        will be linear.
    log_y
        Log base to use for the *y* axis scale, if *None*, the *y* axis scale
        will be linear.
    plot_kwargs
        Keyword arguments for the :func:`matplotlib.pyplot.plot` definition,
        used to control the style of the plotted functions. ``plot_kwargs``
        can be either a single dictionary applied to all the plotted functions
        with the same settings or a sequence of dictionaries with different
        settings for each plotted function.

    Other Parameters
    ----------------
    kwargs
        {:func:`colour.plotting.artist`, :func:`colour.plotting.render`},
        See the documentation of the previously listed definitions.

    Returns
    -------
    :class:`tuple`
        Current figure and axes.

    Examples
    --------
    >>> functions = {
    ...     'Gamma 2.2' : lambda x: x ** (1 / 2.2),
    ...     'Gamma 2.4' : lambda x: x ** (1 / 2.4),
    ...     'Gamma 2.6' : lambda x: x ** (1 / 2.6),
    ... }
    >>> plot_multi_functions(functions)
    ... # doctest: +ELLIPSIS
    (<Figure size ... with 1 Axes>, <...AxesSubplot...>)

    .. image:: ../_static/Plotting_Plot_Multi_Functions.png
        :align: center
        :alt: plot_multi_functions
    """

    settings: Dict[str, Any] = dict(kwargs)

    _figure, axes = artist(**settings)

    plot_settings_collection = [
        {
            "label": f"{name}",
            "zorder": CONSTANTS_COLOUR_STYLE.zorder.midground_label,
        }
        for name in functions.keys()
    ]

    if plot_kwargs is not None:
        update_settings_collection(
            plot_settings_collection, plot_kwargs, len(functions)
        )

    # TODO: Remove when "Matplotlib" minimum version can be set to 3.5.0.
    matplotlib_3_5 = tuple(
        int(token) for token in matplotlib.__version__.split(".")[:2]
    ) >= (3, 5)

    if log_x is not None and log_y is not None:
        attest(
            log_x >= 2 and log_y >= 2,
            "Log base must be equal or greater than 2.",
        )

        plotting_function = axes.loglog

        axes.set_xscale("log", base=log_x)
        axes.set_yscale("log", base=log_y)
    elif log_x is not None:
        attest(log_x >= 2, "Log base must be equal or greater than 2.")

        if matplotlib_3_5:  # pragma: no cover
            plotting_function = partial(axes.semilogx, base=log_x)
        else:  # pragma: no cover
            plotting_function = partial(axes.semilogx, basex=log_x)
    elif log_y is not None:
        attest(log_y >= 2, "Log base must be equal or greater than 2.")

        if matplotlib_3_5:  # pragma: no cover
            plotting_function = partial(axes.semilogy, base=log_y)
        else:  # pragma: no cover
            plotting_function = partial(axes.semilogy, basey=log_y)
    else:
        plotting_function = axes.plot

    samples = cast(ArrayLike, optional(samples, np.linspace(0, 1, 1000)))

    for i, (_name, function) in enumerate(functions.items()):
        plotting_function(
            samples, function(samples), **plot_settings_collection[i]
        )

    x_label = (
        f"x - Log Base {log_x} Scale"
        if log_x is not None
        else "x - Linear Scale"
    )
    y_label = (
        f"y - Log Base {log_y} Scale"
        if log_y is not None
        else "y - Linear Scale"
    )
    settings = {
        "axes": axes,
        "legend": True,
        "title": f"{', '.join(functions)} - Functions",
        "x_label": x_label,
        "y_label": y_label,
    }
    settings.update(kwargs)

    return render(**settings)


@override_style()
def plot_image(
    image: ArrayLike,
    imshow_kwargs: Optional[Dict] = None,
    text_kwargs: Optional[Dict] = None,
    **kwargs: Any,
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Plot given image.

    Parameters
    ----------
    image
        Image to plot.
    imshow_kwargs
        Keyword arguments for the :func:`matplotlib.pyplot.imshow` definition.
    text_kwargs
        Keyword arguments for the :func:`matplotlib.pyplot.text` definition.
        The following special keyword arguments can also be used:

        -   ``offset`` : Sets the text offset.

    Other Parameters
    ----------------
    kwargs
        {:func:`colour.plotting.artist`, :func:`colour.plotting.render`},
        See the documentation of the previously listed definitions.

    Returns
    -------
    :class:`tuple`
        Current figure and axes.

    Examples
    --------
    >>> import os
    >>> import colour
    >>> from colour import read_image
    >>> path = os.path.join(
    ...     colour.__path__[0], 'examples', 'plotting', 'resources',
    ...     'Ishihara_Colour_Blindness_Test_Plate_3.png')
    >>> plot_image(read_image(path))  # doctest: +ELLIPSIS
    (<Figure size ... with 1 Axes>, <...AxesSubplot...>)

    .. image:: ../_static/Plotting_Plot_Image.png
        :align: center
        :alt: plot_image
    """

    _figure, axes = artist(**kwargs)

    imshow_settings = {
        "interpolation": "nearest",
        "cmap": matplotlib.cm.Greys_r,
        "zorder": CONSTANTS_COLOUR_STYLE.zorder.background_polygon,
    }
    if imshow_kwargs is not None:
        imshow_settings.update(imshow_kwargs)

    text_settings = {
        "text": None,
        "offset": 0.005,
        "color": CONSTANTS_COLOUR_STYLE.colour.brightest,
        "alpha": CONSTANTS_COLOUR_STYLE.opacity.high,
        "zorder": CONSTANTS_COLOUR_STYLE.zorder.midground_label,
    }
    if text_kwargs is not None:
        text_settings.update(text_kwargs)
    text_offset = text_settings.pop("offset")

    image = as_float_array(image)

    axes.imshow(np.clip(image, 0, 1), **imshow_settings)

    if text_settings["text"] is not None:
        text = text_settings.pop("text")

        axes.text(
            text_offset,
            text_offset,
            text,
            transform=axes.transAxes,
            ha="left",
            va="bottom",
            **text_settings,
        )

    settings: Dict[str, Any] = {
        "axes": axes,
        "axes_visible": False,
    }
    settings.update(kwargs)

    return render(**settings)



# !/usr/bin/env python
"""Defines the unit tests for the :mod:`colour.appearance.nayatani95` module."""

import numpy as np
import unittest
from itertools import permutations

from colour.appearance import XYZ_to_Nayatani95
from colour.utilities import (
    as_float_array,
    domain_range_scale,
    ignore_numpy_errors,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestXYZ_to_Nayatani95",
]


class TestXYZ_to_Nayatani95(unittest.TestCase):
    """
    Define :func:`colour.appearance.nayatani95.XYZ_to_Nayatani95` definition
    unit tests methods.
    """

    def test_XYZ_to_Nayatani95(self):
        """
        Test :func:`colour.appearance.nayatani95.XYZ_to_Nayatani95`
        definition.

        Notes
        -----
        -   The test values have been generated from data of the following file
            by *Fairchild (2013)*:
            http://rit-mcsl.org/fairchild//files/AppModEx.xls
        """

        XYZ = np.array([19.01, 20.00, 21.78])
        XYZ_n = np.array([95.05, 100.00, 108.88])
        Y_o = 20
        E_o = 5000
        E_or = 1000
        np.testing.assert_allclose(
            XYZ_to_Nayatani95(XYZ, XYZ_n, Y_o, E_o, E_or),
            np.array([50, 0.01, 257.5, 0.01, 62.6, 0.02, np.nan, np.nan, 50]),
            rtol=0.01,
            atol=0.01,
        )

        XYZ = np.array([57.06, 43.06, 31.96])
        E_o = 500
        np.testing.assert_allclose(
            XYZ_to_Nayatani95(XYZ, XYZ_n, Y_o, E_o, E_or),
            np.array([73, 48.3, 21.6, 37.1, 67.3, 42.9, np.nan, np.nan, 75.9]),
            rtol=0.01,
            atol=0.01,
        )

        XYZ = np.array([3.53, 6.56, 2.14])
        XYZ_n = np.array([109.85, 100.00, 35.58])
        E_o = 5000
        np.testing.assert_allclose(
            XYZ_to_Nayatani95(XYZ, XYZ_n, Y_o, E_o, E_or),
            np.array(
                [24.5, 49.3, 190.6, 81.3, 37.5, 62.1, np.nan, np.nan, 29.7]
            ),
            rtol=0.01,
            atol=0.01,
        )

        XYZ = np.array([19.01, 20.00, 21.78])
        E_o = 500
        np.testing.assert_allclose(
            XYZ_to_Nayatani95(XYZ, XYZ_n, Y_o, E_o, E_or),
            np.array(
                [49.4, 39.9, 236.3, 40.2, 44.2, 35.8, np.nan, np.nan, 49.4]
            ),
            rtol=0.01,
            atol=0.01,
        )

    def test_n_dimensional_XYZ_to_Nayatani95(self):
        """
        Test :func:`colour.appearance.nayatani95.XYZ_to_Nayatani95` definition
        n-dimensional support.
        """

        XYZ = np.array([19.01, 20.00, 21.78])
        XYZ_n = np.array([95.05, 100.00, 108.88])
        Y_o = 20
        E_o = 5000
        E_or = 1000
        specification = XYZ_to_Nayatani95(XYZ, XYZ_n, Y_o, E_o, E_or)

        XYZ = np.tile(XYZ, (6, 1))
        specification = np.tile(specification, (6, 1))
        np.testing.assert_almost_equal(
            XYZ_to_Nayatani95(XYZ, XYZ_n, Y_o, E_o, E_or),
            specification,
            decimal=7,
        )

        XYZ_n = np.tile(XYZ_n, (6, 1))
        np.testing.assert_almost_equal(
            XYZ_to_Nayatani95(XYZ, XYZ_n, Y_o, E_o, E_or),
            specification,
            decimal=7,
        )

        XYZ = np.reshape(XYZ, (2, 3, 3))
        XYZ_n = np.reshape(XYZ_n, (2, 3, 3))
        specification = np.reshape(specification, (2, 3, 9))
        np.testing.assert_almost_equal(
            XYZ_to_Nayatani95(XYZ, XYZ_n, Y_o, E_o, E_or),
            specification,
            decimal=7,
        )

    def test_domain_range_scale_XYZ_to_Nayatani95(self):
        """
        Test :func:`colour.appearance.nayatani95.XYZ_to_Nayatani95` definition
        domain and range scale support.
        """

        XYZ = np.array([19.01, 20.00, 21.78])
        XYZ_n = np.array([95.05, 100.00, 108.88])
        Y_o = 20.0
        E_o = 5000.0
        E_or = 1000.0
        specification = XYZ_to_Nayatani95(XYZ, XYZ_n, Y_o, E_o, E_or)

        d_r = (
            ("reference", 1, 1),
            ("1", 0.01, np.array([1, 1, 1 / 360, 1, 1, 1, np.nan, np.nan, 1])),
            (
                "100",
                1,
                np.array([1, 1, 100 / 360, 1, 1, 1, np.nan, np.nan, 1]),
            ),
        )
        for scale, factor_a, factor_b in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    XYZ_to_Nayatani95(
                        XYZ * factor_a, XYZ_n * factor_a, Y_o, E_o, E_or
                    ),
                    as_float_array(specification) * factor_b,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_XYZ_to_Nayatani95(self):
        """
        Test :func:`colour.appearance.nayatani95.XYZ_to_Nayatani95` definition
        nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            XYZ = np.array(case)
            XYZ_n = np.array(case)
            Y_o = case[0]
            E_o = case[0]
            E_or = case[0]
            XYZ_to_Nayatani95(XYZ, XYZ_n, Y_o, E_o, E_or)



from .hunt import (
    InductionFactors_Hunt,
    VIEWING_CONDITIONS_HUNT,
    CAM_Specification_Hunt,
    XYZ_to_Hunt,
)
from .atd95 import CAM_Specification_ATD95, XYZ_to_ATD95
from .ciecam02 import (
    InductionFactors_CIECAM02,
    VIEWING_CONDITIONS_CIECAM02,
    CAM_KWARGS_CIECAM02_sRGB,
    CAM_Specification_CIECAM02,
    XYZ_to_CIECAM02,
    CIECAM02_to_XYZ,
)
from .cam16 import (
    InductionFactors_CAM16,
    VIEWING_CONDITIONS_CAM16,
    CAM_Specification_CAM16,
    XYZ_to_CAM16,
    CAM16_to_XYZ,
)
from .hke import (
    HKE_NAYATANI1997_METHODS,
    HelmholtzKohlrausch_effect_object_Nayatani1997,
    HelmholtzKohlrausch_effect_luminous_Nayatani1997,
)
from .hke import coefficient_q_Nayatani1997, coefficient_K_Br_Nayatani1997
from .kim2009 import (
    InductionFactors_Kim2009,
    VIEWING_CONDITIONS_KIM2009,
    MediaParameters_Kim2009,
    MEDIA_PARAMETERS_KIM2009,
    CAM_Specification_Kim2009,
    XYZ_to_Kim2009,
    Kim2009_to_XYZ,
)
from .llab import (
    InductionFactors_LLAB,
    VIEWING_CONDITIONS_LLAB,
    CAM_Specification_LLAB,
    XYZ_to_LLAB,
)
from .nayatani95 import CAM_Specification_Nayatani95, XYZ_to_Nayatani95
from .rlab import (
    VIEWING_CONDITIONS_RLAB,
    D_FACTOR_RLAB,
    CAM_Specification_RLAB,
    XYZ_to_RLAB,
)
from .zcam import (
    InductionFactors_ZCAM,
    VIEWING_CONDITIONS_ZCAM,
    CAM_Specification_ZCAM,
    XYZ_to_ZCAM,
    ZCAM_to_XYZ,
)

__all__ = [
    "InductionFactors_Hunt",
    "VIEWING_CONDITIONS_HUNT",
    "CAM_Specification_Hunt",
    "XYZ_to_Hunt",
]
__all__ += [
    "CAM_Specification_ATD95",
    "XYZ_to_ATD95",
]
__all__ += [
    "InductionFactors_CIECAM02",
    "VIEWING_CONDITIONS_CIECAM02",
    "CAM_KWARGS_CIECAM02_sRGB",
    "CAM_Specification_CIECAM02",
    "XYZ_to_CIECAM02",
    "CIECAM02_to_XYZ",
]
__all__ += [
    "InductionFactors_CAM16",
    "VIEWING_CONDITIONS_CAM16",
    "CAM_Specification_CAM16",
    "XYZ_to_CAM16",
    "CAM16_to_XYZ",
]
__all__ += [
    "HKE_NAYATANI1997_METHODS",
    "HelmholtzKohlrausch_effect_object_Nayatani1997",
    "HelmholtzKohlrausch_effect_luminous_Nayatani1997",
]
__all__ += [
    "coefficient_q_Nayatani1997",
    "coefficient_K_Br_Nayatani1997",
]
__all__ += [
    "InductionFactors_Kim2009",
    "VIEWING_CONDITIONS_KIM2009",
    "MediaParameters_Kim2009",
    "MEDIA_PARAMETERS_KIM2009",
    "CAM_Specification_Kim2009",
    "XYZ_to_Kim2009",
    "Kim2009_to_XYZ",
]
__all__ += [
    "InductionFactors_LLAB",
    "VIEWING_CONDITIONS_LLAB",
    "CAM_Specification_LLAB",
    "XYZ_to_LLAB",
]
__all__ += [
    "CAM_Specification_Nayatani95",
    "XYZ_to_Nayatani95",
]
__all__ += [
    "VIEWING_CONDITIONS_RLAB",
    "D_FACTOR_RLAB",
    "CAM_Specification_RLAB",
    "XYZ_to_RLAB",
]
__all__ += [
    "InductionFactors_ZCAM",
    "VIEWING_CONDITIONS_ZCAM",
    "CAM_Specification_ZCAM",
    "XYZ_to_ZCAM",
    "ZCAM_to_XYZ",
]



"""
:math:`IC_TC_P` Colour Encoding
===============================

Defines the :math:`IC_TC_P` colour encoding related transformations:

-   :func:`colour.RGB_to_ICtCp`
-   :func:`colour.ICtCp_to_RGB`
-   :func:`colour.XYZ_to_ICtCp`
-   :func:`colour.ICtCp_to_XYZ`

References
----------
-   :cite:`Dolby2016a` : Dolby. (2016). WHAT IS ICtCp? - INTRODUCTION.
    https://www.dolby.com/us/en/technologies/dolby-vision/ICtCp-white-paper.pdf
-   :cite:`InternationalTelecommunicationUnion2018` : International
    Telecommunication Union. (2018). Recommendation ITU-R BT.2100-2 - Image
    parameter values for high dynamic range television for use in production
    and international programme exchange.
    https://www.itu.int/dms_pubrec/itu-r/rec/bt/\
R-REC-BT.2100-2-201807-I!!PDF-E.pdf
-   :cite:`Lu2016c` : Lu, T., Pu, F., Yin, P., Chen, T., Husak, W., Pytlarz,
    J., Atkins, R., Froehlich, J., & Su, G.-M. (2016). ITP Colour Space and Its
    Compression Performance for High Dynamic Range and Wide Colour Gamut Video
    Distribution. ZTE Communications, 14(1), 32-38.
"""

from __future__ import annotations

import numpy as np

from colour.algebra import vector_dot
from colour.colorimetry import CCS_ILLUMINANTS
from colour.hints import ArrayLike, Floating, Literal, NDArray, Union
from colour.models.rgb import RGB_COLOURSPACES, RGB_to_XYZ, XYZ_to_RGB
from colour.models.rgb.transfer_functions import (
    eotf_ST2084,
    eotf_inverse_ST2084,
    oetf_HLG_BT2100,
    oetf_inverse_HLG_BT2100,
)
from colour.utilities import (
    as_float_array,
    domain_range_scale,
    validate_method,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "MATRIX_ICTCP_RGB_TO_LMS",
    "MATRIX_ICTCP_LMS_TO_RGB",
    "MATRIX_ICTCP_LMS_P_TO_ICTCP",
    "MATRIX_ICTCP_ICTCP_TO_LMS_P",
    "MATRIX_ICTCP_LMS_P_TO_ICTCP_HLG_BT2100_2",
    "MATRIX_ICTCP_ICTCP_TO_LMS_P_HLG_BT2100_2",
    "RGB_to_ICtCp",
    "ICtCp_to_RGB",
    "XYZ_to_ICtCp",
    "ICtCp_to_XYZ",
]

MATRIX_ICTCP_RGB_TO_LMS: NDArray = (
    np.array(
        [
            [1688, 2146, 262],
            [683, 2951, 462],
            [99, 309, 3688],
        ]
    )
    / 4096
)
"""*ITU-R BT.2020* colourspace to normalised cone responses matrix."""

MATRIX_ICTCP_LMS_TO_RGB: NDArray = np.linalg.inv(MATRIX_ICTCP_RGB_TO_LMS)
"""
:math:`IC_TC_P` colourspace normalised cone responses to *ITU-R BT.2020*
colourspace matrix.
"""

MATRIX_ICTCP_LMS_P_TO_ICTCP: NDArray = (
    np.array(
        [
            [2048, 2048, 0],
            [6610, -13613, 7003],
            [17933, -17390, -543],
        ]
    )
    / 4096
)
"""
:math:`LMS_p` *SMPTE ST 2084:2014* encoded normalised cone responses to
:math:`IC_TC_P` colour encoding matrix.
"""

MATRIX_ICTCP_ICTCP_TO_LMS_P: NDArray = np.linalg.inv(
    MATRIX_ICTCP_LMS_P_TO_ICTCP
)
"""
:math:`IC_TC_P` colour encoding to :math:`LMS_p` *SMPTE ST 2084:2014* encoded
normalised cone responses matrix.
"""

MATRIX_ICTCP_LMS_P_TO_ICTCP_HLG_BT2100_2: NDArray = (
    np.array(
        [
            [2048, 2048, 0],
            [3625, -7465, 3840],
            [9500, -9212, -288],
        ]
    )
    / 4096
)
"""
:math:`LMS_p` *SMPTE ST 2084:2014* encoded normalised cone responses to
:math:`IC_TC_P` colour encoding matrix as given in *ITU-R BT.2100-2*.
"""

MATRIX_ICTCP_ICTCP_TO_LMS_P_HLG_BT2100_2: NDArray = np.linalg.inv(
    MATRIX_ICTCP_LMS_P_TO_ICTCP_HLG_BT2100_2
)
"""
:math:`IC_TC_P` colour encoding to :math:`LMS_p` *SMPTE ST 2084:2014* encoded
normalised cone responses matrix as given in *ITU-R BT.2100-2*.
"""


def RGB_to_ICtCp(
    RGB: ArrayLike,
    method: Union[
        Literal[
            "Dolby 2016",
            "ITU-R BT.2100-1 HLG",
            "ITU-R BT.2100-1 PQ",
            "ITU-R BT.2100-2 HLG",
            "ITU-R BT.2100-2 PQ",
        ],
        str,
    ] = "Dolby 2016",
    L_p: Floating = 10000,
) -> NDArray:
    """
    Convert from *ITU-R BT.2020* colourspace to :math:`IC_TC_P` colour
    encoding.

    Parameters
    ----------
    RGB
        *ITU-R BT.2020* colourspace array.
    method
        Computation method. *Recommendation ITU-R BT.2100* defines multiple
        variants of the :math:`IC_TC_P` colour encoding:

        -   *ITU-R BT.2100-1*

            -   *SMPTE ST 2084:2014* inverse electro-optical transfer
                function (EOTF) and the :math:`IC_TC_P` matrix from
                :cite:`Dolby2016a`: *Dolby 2016*, *ITU-R BT.2100-1 PQ*,
                *ITU-R BT.2100-2 PQ* methods.
            -   *Recommendation ITU-R BT.2100* *Reference HLG* opto-electrical
                transfer function (OETF) and the :math:`IC_TC_P` matrix
                from :cite:`Dolby2016a`: *ITU-R BT.2100-1 HLG* method.

        -   *ITU-R BT.2100-2*

            -   *SMPTE ST 2084:2014* inverse electro-optical transfer
                function (EOTF) and the :math:`IC_TC_P` matrix from
                :cite:`Dolby2016a`: *Dolby 2016*, *ITU-R BT.2100-1 PQ*,
                *ITU-R BT.2100-2 PQ* methods.
            -   *Recommendation ITU-R BT.2100* *Reference HLG* opto-electrical
                transfer function (OETF) and a custom :math:`IC_TC_P`
                matrix from :cite:`InternationalTelecommunicationUnion2018`:
                *ITU-R BT.2100-2 HLG* method.

    L_p
        Display peak luminance :math:`cd/m^2` for *SMPTE ST 2084:2014*
        non-linear encoding. This parameter should stay at its default
        :math:`10000 cd/m^2` value for practical applications. It is exposed so
        that the definition can be used as a fitting function.

    Returns
    -------
    :class:`numpy.ndarray`
        :math:`IC_TC_P` colour encoding array.

    Warnings
    --------
    The underlying *SMPTE ST 2084:2014* transfer function is an absolute
    transfer function.

    Notes
    -----
    -   The *ITU-R BT.2100-1 PQ* and *ITU-R BT.2100-2 PQ* methods are aliases
        for the *Dolby 2016* method.
    -   The underlying *SMPTE ST 2084:2014* transfer function is an absolute
        transfer function, thus the domain and range values for the *Reference*
        and *1* scales are only indicative that the data is not affected by
        scale transformations. The effective domain of *SMPTE ST 2084:2014*
        inverse electro-optical transfer function (EOTF) is
        [0.0001, 10000].

    +------------+-----------------------+------------------+
    | **Domain** | **Scale - Reference** | **Scale - 1**    |
    +============+=======================+==================+
    | ``RGB``    | ``UN``                | ``UN``           |
    +------------+-----------------------+------------------+

    +------------+-----------------------+------------------+
    | **Range**  | **Scale - Reference** | **Scale - 1**    |
    +============+=======================+==================+
    | ``ICtCp``  | ``I``  : [0, 1]       | ``I``  : [0, 1]  |
    |            |                       |                  |
    |            | ``CT`` : [-1, 1]      | ``CT`` : [-1, 1] |
    |            |                       |                  |
    |            | ``CP`` : [-1, 1]      | ``CP`` : [-1, 1] |
    +------------+-----------------------+------------------+

    References
    ----------
    :cite:`Dolby2016a`, :cite:`Lu2016c`

    Examples
    --------
    >>> RGB = np.array([0.45620519, 0.03081071, 0.04091952])
    >>> RGB_to_ICtCp(RGB)  # doctest: +ELLIPSIS
    array([ 0.0735136...,  0.0047525...,  0.0935159...])
    >>> RGB_to_ICtCp(RGB, method='ITU-R BT.2100-2 HLG')  # doctest: +ELLIPSIS
    array([ 0.6256789..., -0.0198449...,  0.3591125...])
    """

    RGB = as_float_array(RGB)
    method = validate_method(
        method,
        [
            "Dolby 2016",
            "ITU-R BT.2100-1 HLG",
            "ITU-R BT.2100-1 PQ",
            "ITU-R BT.2100-2 HLG",
            "ITU-R BT.2100-2 PQ",
        ],
    )

    is_hlg_method = "hlg" in method
    is_BT2100_2_method = "2100-2" in method

    LMS = vector_dot(MATRIX_ICTCP_RGB_TO_LMS, RGB)

    with domain_range_scale("ignore"):
        LMS_p = (
            oetf_HLG_BT2100(LMS)
            if is_hlg_method
            else eotf_inverse_ST2084(LMS, L_p)
        )

    ICtCp = (
        vector_dot(MATRIX_ICTCP_LMS_P_TO_ICTCP_HLG_BT2100_2, LMS_p)
        if (is_hlg_method and is_BT2100_2_method)
        else vector_dot(MATRIX_ICTCP_LMS_P_TO_ICTCP, LMS_p)
    )

    return ICtCp


def ICtCp_to_RGB(
    ICtCp: ArrayLike,
    method: Union[
        Literal[
            "Dolby 2016",
            "ITU-R BT.2100-1 HLG",
            "ITU-R BT.2100-1 PQ",
            "ITU-R BT.2100-2 HLG",
            "ITU-R BT.2100-2 PQ",
        ],
        str,
    ] = "Dolby 2016",
    L_p: Floating = 10000,
) -> NDArray:
    """
    Convert from :math:`IC_TC_P` colour encoding to *ITU-R BT.2020*
    colourspace.

    Parameters
    ----------
    ICtCp
        :math:`IC_TC_P` colour encoding array.
    method
        Computation method. *Recommendation ITU-R BT.2100* defines multiple
        variants of the :math:`IC_TC_P` colour encoding:

        -   *ITU-R BT.2100-1*

            -   *SMPTE ST 2084:2014* inverse electro-optical transfer
                function (EOTF) and the :math:`IC_TC_P` matrix from
                :cite:`Dolby2016a`: *Dolby 2016*, *ITU-R BT.2100-1 PQ*,
                *ITU-R BT.2100-2 PQ* methods.
            -   *Recommendation ITU-R BT.2100* *Reference HLG* opto-electrical
                transfer function (OETF) and the :math:`IC_TC_P` matrix
                from :cite:`Dolby2016a`: *ITU-R BT.2100-1 HLG* method.

        -   *ITU-R BT.2100-2*

            -   *SMPTE ST 2084:2014* inverse electro-optical transfer
                function (EOTF) and the :math:`IC_TC_P` matrix from
                :cite:`Dolby2016a`: *Dolby 2016*, *ITU-R BT.2100-1 PQ*,
                *ITU-R BT.2100-2 PQ* methods.
            -   *Recommendation ITU-R BT.2100* *Reference HLG* opto-electrical
                transfer function (OETF) and a custom :math:`IC_TC_P`
                matrix from :cite:`InternationalTelecommunicationUnion2018`:
                *ITU-R BT.2100-2 HLG* method.

    L_p
        Display peak luminance :math:`cd/m^2` for *SMPTE ST 2084:2014*
        non-linear encoding. This parameter should stay at its default
        :math:`10000 cd/m^2` value for practical applications. It is exposed so
        that the definition can be used as a fitting function.

    Returns
    -------
    :class:`numpy.ndarray`
        *ITU-R BT.2020* colourspace array.

    Warnings
    --------
    The underlying *SMPTE ST 2084:2014* transfer function is an absolute
    transfer function.

    Notes
    -----
    -   The *ITU-R BT.2100-1 PQ* and *ITU-R BT.2100-2 PQ* methods are aliases
        for the *Dolby 2016* method.
    -   The underlying *SMPTE ST 2084:2014* transfer function is an absolute
        transfer function, thus the domain and range values for the *Reference*
        and *1* scales are only indicative that the data is not affected by
        scale transformations.

    +------------+-----------------------+------------------+
    | **Domain** | **Scale - Reference** | **Scale - 1**    |
    +============+=======================+==================+
    | ``ICtCp``  | ``I``  : [0, 1]       | ``I``  : [0, 1]  |
    |            |                       |                  |
    |            | ``CT`` : [-1, 1]      | ``CT`` : [-1, 1] |
    |            |                       |                  |
    |            | ``CP`` : [-1, 1]      | ``CP`` : [-1, 1] |
    +------------+-----------------------+------------------+

    +------------+-----------------------+------------------+
    | **Range**  | **Scale - Reference** | **Scale - 1**    |
    +============+=======================+==================+
    | ``RGB``    | ``UN``                | ``UN``           |
    +------------+-----------------------+------------------+

    References
    ----------
    :cite:`Dolby2016a`, :cite:`Lu2016c`

    Examples
    --------
    >>> ICtCp = np.array([0.07351364, 0.00475253, 0.09351596])
    >>> ICtCp_to_RGB(ICtCp)  # doctest: +ELLIPSIS
    array([ 0.4562052...,  0.0308107...,  0.0409195...])
    >>> ICtCp = np.array([0.62567899, -0.01984490, 0.35911259])
    >>> ICtCp_to_RGB(ICtCp, method='ITU-R BT.2100-2 HLG')  # doctest: +ELLIPSIS
    array([ 0.4562052...,  0.0308107...,  0.0409195...])
    """

    ICtCp = as_float_array(ICtCp)
    method = validate_method(
        method,
        [
            "Dolby 2016",
            "ITU-R BT.2100-1 HLG",
            "ITU-R BT.2100-1 PQ",
            "ITU-R BT.2100-2 HLG",
            "ITU-R BT.2100-2 PQ",
        ],
    )

    is_hlg_method = "hlg" in method
    is_BT2100_2_method = "2100-2" in method

    LMS_p = (
        vector_dot(MATRIX_ICTCP_ICTCP_TO_LMS_P_HLG_BT2100_2, ICtCp)
        if (is_hlg_method and is_BT2100_2_method)
        else vector_dot(MATRIX_ICTCP_ICTCP_TO_LMS_P, ICtCp)
    )

    with domain_range_scale("ignore"):
        LMS = (
            oetf_inverse_HLG_BT2100(LMS_p)
            if is_hlg_method
            else eotf_ST2084(LMS_p, L_p)
        )

    RGB = vector_dot(MATRIX_ICTCP_LMS_TO_RGB, LMS)

    return RGB


def XYZ_to_ICtCp(
    XYZ: ArrayLike,
    illuminant=CCS_ILLUMINANTS["CIE 1931 2 Degree Standard Observer"]["D65"],
    chromatic_adaptation_transform: Union[
        Literal[
            "Bianco 2010",
            "Bianco PC 2010",
            "Bradford",
            "CAT02 Brill 2008",
            "CAT02",
            "CAT16",
            "CMCCAT2000",
            "CMCCAT97",
            "Fairchild",
            "Sharp",
            "Von Kries",
            "XYZ Scaling",
        ],
        str,
    ] = "CAT02",
    method: Union[
        Literal[
            "Dolby 2016",
            "ITU-R BT.2100-1 HLG",
            "ITU-R BT.2100-1 PQ",
            "ITU-R BT.2100-2 HLG",
            "ITU-R BT.2100-2 PQ",
        ],
        str,
    ] = "Dolby 2016",
    L_p: Floating = 10000,
) -> NDArray:
    """
    Convert from *CIE XYZ* tristimulus values to :math:`IC_TC_P` colour
    encoding.

    Parameters
    ----------
    XYZ
        *CIE XYZ* tristimulus values.
    illuminant
        Source illuminant chromaticity coordinates.
    chromatic_adaptation_transform
        *Chromatic adaptation* transform.
    method
        Computation method. *Recommendation ITU-R BT.2100* defines multiple
        variants of the :math:`IC_TC_P` colour encoding:

        -   *ITU-R BT.2100-1*

            -   *SMPTE ST 2084:2014* inverse electro-optical transfer
                function (EOTF) and the :math:`IC_TC_P` matrix from
                :cite:`Dolby2016a`: *Dolby 2016*, *ITU-R BT.2100-1 PQ*,
                *ITU-R BT.2100-2 PQ* methods.
            -   *Recommendation ITU-R BT.2100* *Reference HLG* opto-electrical
                transfer function (OETF) and the :math:`IC_TC_P` matrix
                from :cite:`Dolby2016a`: *ITU-R BT.2100-1 HLG* method.

        -   *ITU-R BT.2100-2*

            -   *SMPTE ST 2084:2014* inverse electro-optical transfer
                function (EOTF) and the :math:`IC_TC_P` matrix from
                :cite:`Dolby2016a`: *Dolby 2016*, *ITU-R BT.2100-1 PQ*,
                *ITU-R BT.2100-2 PQ* methods.
            -   *Recommendation ITU-R BT.2100* *Reference HLG* opto-electrical
                transfer function (OETF) and a custom :math:`IC_TC_P`
                matrix from :cite:`InternationalTelecommunicationUnion2018`:
                *ITU-R BT.2100-2 HLG* method.

    L_p
        Display peak luminance :math:`cd/m^2` for *SMPTE ST 2084:2014*
        non-linear encoding. This parameter should stay at its default
        :math:`10000 cd/m^2` value for practical applications. It is exposed so
        that the definition can be used as a fitting function.

    Returns
    -------
    :class:`numpy.ndarray`
        :math:`IC_TC_P` colour encoding array.

    Warnings
    --------
    The underlying *SMPTE ST 2084:2014* transfer function is an absolute
    transfer function.

    Notes
    -----
    -   The underlying *SMPTE ST 2084:2014* transfer function is an absolute
        transfer function, thus the domain and range values for the *Reference*
    -   The *ITU-R BT.2100-1 PQ* and *ITU-R BT.2100-2 PQ* methods are aliases
        for the *Dolby 2016* method.
        and *1* scales are only indicative that the data is not affected by
        scale transformations. The effective domain of *SMPTE ST 2084:2014*
        inverse electro-optical transfer function (EOTF) is
        [0.0001, 10000].

    +------------+-----------------------+------------------+
    | **Domain** | **Scale - Reference** | **Scale - 1**    |
    +============+=======================+==================+
    | ``XYZ``    | ``UN``                | ``UN``           |
    +------------+-----------------------+------------------+

    +------------+-----------------------+------------------+
    | **Range**  | **Scale - Reference** | **Scale - 1**    |
    +============+=======================+==================+
    | ``ICtCp``  | ``I``  : [0, 1]       | ``I``  : [0, 1]  |
    |            |                       |                  |
    |            | ``CT`` : [-1, 1]      | ``CT`` : [-1, 1] |
    |            |                       |                  |
    |            | ``CP`` : [-1, 1]      | ``CP`` : [-1, 1] |
    +------------+-----------------------+------------------+

    References
    ----------
    :cite:`Dolby2016a`, :cite:`Lu2016c`

    Examples
    --------
    >>> XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
    >>> XYZ_to_ICtCp(XYZ)  # doctest: +ELLIPSIS
    array([ 0.0685809..., -0.0028384...,  0.0602098...])
    >>> XYZ_to_ICtCp(XYZ, method='ITU-R BT.2100-2 HLG')  # doctest: +ELLIPSIS
    array([ 0.5924279..., -0.0374073...,  0.2512267...])
    """

    BT2020 = RGB_COLOURSPACES["ITU-R BT.2020"]

    RGB = XYZ_to_RGB(
        XYZ,
        illuminant,
        BT2020.whitepoint,
        BT2020.matrix_XYZ_to_RGB,
        chromatic_adaptation_transform,
    )

    return RGB_to_ICtCp(RGB, method, L_p)


def ICtCp_to_XYZ(
    ICtCp: ArrayLike,
    illuminant=CCS_ILLUMINANTS["CIE 1931 2 Degree Standard Observer"]["D65"],
    chromatic_adaptation_transform: Union[
        Literal[
            "Bianco 2010",
            "Bianco PC 2010",
            "Bradford",
            "CAT02 Brill 2008",
            "CAT02",
            "CAT16",
            "CMCCAT2000",
            "CMCCAT97",
            "Fairchild",
            "Sharp",
            "Von Kries",
            "XYZ Scaling",
        ],
        str,
    ] = "CAT02",
    method: Union[
        Literal[
            "Dolby 2016",
            "ITU-R BT.2100-1 HLG",
            "ITU-R BT.2100-1 PQ",
            "ITU-R BT.2100-2 HLG",
            "ITU-R BT.2100-2 PQ",
        ],
        str,
    ] = "Dolby 2016",
    L_p: Floating = 10000,
) -> NDArray:
    """
    Convert from :math:`IC_TC_P` colour encoding to *CIE XYZ* tristimulus
    values.

    Parameters
    ----------
    ICtCp
        :math:`IC_TC_P` colour encoding array.
    illuminant
        Source illuminant chromaticity coordinates.
    chromatic_adaptation_transform
        *Chromatic adaptation* transform.
    method
        Computation method. *Recommendation ITU-R BT.2100* defines multiple
        variants of the :math:`IC_TC_P` colour encoding:

        -   *ITU-R BT.2100-1*

            -   *SMPTE ST 2084:2014* inverse electro-optical transfer
                function (EOTF) and the :math:`IC_TC_P` matrix from
                :cite:`Dolby2016a`: *Dolby 2016*, *ITU-R BT.2100-1 PQ*,
                *ITU-R BT.2100-2 PQ* methods.
            -   *Recommendation ITU-R BT.2100* *Reference HLG* opto-electrical
                transfer function (OETF) and the :math:`IC_TC_P` matrix
                from :cite:`Dolby2016a`: *ITU-R BT.2100-1 HLG* method.

        -   *ITU-R BT.2100-2*

            -   *SMPTE ST 2084:2014* inverse electro-optical transfer
                function (EOTF) and the :math:`IC_TC_P` matrix from
                :cite:`Dolby2016a`: *Dolby 2016*, *ITU-R BT.2100-1 PQ*,
                *ITU-R BT.2100-2 PQ* methods.
            -   *Recommendation ITU-R BT.2100* *Reference HLG* opto-electrical
                transfer function (OETF) and a custom :math:`IC_TC_P`
                matrix from :cite:`InternationalTelecommunicationUnion2018`:
                *ITU-R BT.2100-2 HLG* method.

    L_p
        Display peak luminance :math:`cd/m^2` for *SMPTE ST 2084:2014*
        non-linear encoding. This parameter should stay at its default
        :math:`10000 cd/m^2` value for practical applications. It is exposed so
        that the definition can be used as a fitting function.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE XYZ* tristimulus values.

    Warnings
    --------
    The underlying *SMPTE ST 2084:2014* transfer function is an absolute
    transfer function.

    Notes
    -----
    -   The *ITU-R BT.2100-1 PQ* and *ITU-R BT.2100-2 PQ* methods are aliases
        for the *Dolby 2016* method.
    -   The underlying *SMPTE ST 2084:2014* transfer function is an absolute
        transfer function, thus the domain and range values for the *Reference*
        and *1* scales are only indicative that the data is not affected by
        scale transformations.

    +------------+-----------------------+------------------+
    | **Domain** | **Scale - Reference** | **Scale - 1**    |
    +============+=======================+==================+
    | ``ICtCp``  | ``I``  : [0, 1]       | ``I``  : [0, 1]  |
    |            |                       |                  |
    |            | ``CT`` : [-1, 1]      | ``CT`` : [-1, 1] |
    |            |                       |                  |
    |            | ``CP`` : [-1, 1]      | ``CP`` : [-1, 1] |
    +------------+-----------------------+------------------+

    +------------+-----------------------+------------------+
    | **Range**  | **Scale - Reference** | **Scale - 1**    |
    +============+=======================+==================+
    | ``XYZ``    | ``UN``                | ``UN``           |
    +------------+-----------------------+------------------+

    References
    ----------
    :cite:`Dolby2016a`, :cite:`Lu2016c`

    Examples
    --------
    >>> ICtCp = np.array([0.06858097, -0.00283842, 0.06020983])
    >>> ICtCp_to_XYZ(ICtCp)  # doctest: +ELLIPSIS
    array([ 0.2065400...,  0.1219722...,  0.0513695...])
    >>> ICtCp = np.array([0.59242792, -0.03740730, 0.25122675])
    >>> ICtCp_to_XYZ(ICtCp, method='ITU-R BT.2100-2 HLG')  # doctest: +ELLIPSIS
    array([ 0.2065400...,  0.1219722...,  0.0513695...])
    """

    RGB = ICtCp_to_RGB(ICtCp, method, L_p)

    BT2020 = RGB_COLOURSPACES["ITU-R BT.2020"]

    XYZ = RGB_to_XYZ(
        RGB,
        BT2020.whitepoint,
        illuminant,
        BT2020.matrix_RGB_to_XYZ,
        chromatic_adaptation_transform,
    )

    return XYZ



"""
LUT Operator
============

Defines the *LUT* sequence class:

-   :class:`colour.LUTSequence`
"""

from __future__ import annotations

import re
from collections.abc import MutableSequence
from copy import deepcopy

from colour.hints import (
    Any,
    ArrayLike,
    Integer,
    List,
    NDArray,
    Sequence,
    TypeLUTSequenceItem,
    Union,
)
from colour.utilities import as_float_array, attest, is_iterable

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "LUTSequence",
]


class LUTSequence(MutableSequence):
    """
    Define the base class for a *LUT* sequence, i.e. a series of *LUTs*,
    *LUT* operators or objects implementing the
    :class:`colour.hints.TypeLUTSequenceItem` protocol.

    The `colour.LUTSequence` class can be used to model series of *LUTs* such
    as when a shaper *LUT* is combined with a 3D *LUT*.

    Other Parameters
    ----------------
    args
        Sequence of objects implementing the
        :class:`colour.hints.TypeLUTSequenceItem` protocol.

    Attributes
    ----------
    -   :attr:`~colour.LUTSequence.sequence`

    Methods
    -------
    -   :meth:`~colour.LUTSequence.__init__`
    -   :meth:`~colour.LUTSequence.__getitem__`
    -   :meth:`~colour.LUTSequence.__setitem__`
    -   :meth:`~colour.LUTSequence.__delitem__`
    -   :meth:`~colour.LUTSequence.__len__`
    -   :meth:`~colour.LUTSequence.__str__`
    -   :meth:`~colour.LUTSequence.__repr__`
    -   :meth:`~colour.LUTSequence.__eq__`
    -   :meth:`~colour.LUTSequence.__ne__`
    -   :meth:`~colour.LUTSequence.insert`
    -   :meth:`~colour.LUTSequence.apply`
    -   :meth:`~colour.LUTSequence.copy`

    Examples
    --------
    >>> from colour.io.luts import  LUT1D, LUT3x1D, LUT3D
    >>> LUT_1 = LUT1D()
    >>> LUT_2 = LUT3D(size=3)
    >>> LUT_3 = LUT3x1D()
    >>> print(LUTSequence(LUT_1, LUT_2, LUT_3))
    LUT Sequence
    ------------
    <BLANKLINE>
    Overview
    <BLANKLINE>
        LUT1D --> LUT3D --> LUT3x1D
    <BLANKLINE>
    Operations
    <BLANKLINE>
        LUT1D - Unity 10
        ----------------
    <BLANKLINE>
        Dimensions : 1
        Domain     : [ 0.  1.]
        Size       : (10,)
    <BLANKLINE>
        LUT3D - Unity 3
        ---------------
    <BLANKLINE>
        Dimensions : 3
        Domain     : [[ 0.  0.  0.]
                      [ 1.  1.  1.]]
        Size       : (3, 3, 3, 3)
    <BLANKLINE>
        LUT3x1D - Unity 10
        ------------------
    <BLANKLINE>
        Dimensions : 2
        Domain     : [[ 0.  0.  0.]
                      [ 1.  1.  1.]]
        Size       : (10, 3)
    """

    def __init__(self, *args: TypeLUTSequenceItem):
        # TODO: Remove pragma when https://github.com/python/mypy/issues/3004
        # is resolved.
        self._sequence: List[TypeLUTSequenceItem] = []
        self.sequence = args  # type: ignore[assignment]

    @property
    def sequence(self) -> List[TypeLUTSequenceItem]:
        """
        Getter and setter property for the underlying *LUT* sequence.

        Parameters
        ----------
        value
            Value to set the underlying *LUT* sequence with.

        Returns
        -------
        :class:`list`
            Underlying *LUT* sequence.
        """

        return self._sequence

    @sequence.setter
    def sequence(self, value: Sequence[TypeLUTSequenceItem]):
        """Setter for the **self.sequence** property."""

        for item in value:
            attest(
                isinstance(item, TypeLUTSequenceItem),
                '"value" items must implement the "TypeLUTSequenceItem" '
                "protocol!",
            )

        self._sequence = list(value)

    def __getitem__(self, index: Union[Integer, slice]) -> Any:
        """
        Return the *LUT* sequence item(s) at given index (or slice).

        Parameters
        ----------
        index
            Index (or slice) to return the *LUT* sequence item(s) at.

        Returns
        -------
        TypeLUTSequenceItem
            *LUT* sequence item(s) at given index (or slice).
        """

        return self._sequence[index]

    def __setitem__(self, index: Union[Integer, slice], value: Any):
        """
        Set the *LUT* sequence at given index (or slice) with given value.

        Parameters
        ----------
        index
            Index (or slice) to set the *LUT* sequence value at.
        value
            Value to set the *LUT* sequence with.
        """

        for item in value if is_iterable(value) else [value]:
            attest(
                isinstance(item, TypeLUTSequenceItem),
                '"value" items must implement the "TypeLUTSequenceItem" '
                "protocol!",
            )

        self._sequence[index] = value

    def __delitem__(self, index: Union[Integer, slice]):
        """
        Delete the *LUT* sequence item(s) at given index (or slice).

        Parameters
        ----------
        index
            Index (or slice) to delete the *LUT* sequence items at.
        """

        del self._sequence[index]

    def __len__(self) -> Integer:
        """
        Return the *LUT* sequence items count.

        Returns
        -------
        :class:`numpy.integer`
            *LUT* sequence items count.
        """

        return len(self._sequence)

    def __str__(self) -> str:
        """
        Return a formatted string representation of the *LUT* sequence.

        Returns
        -------
        :class:`str`
            Formatted string representation.
        """

        sequence = " --> ".join([a.__class__.__name__ for a in self._sequence])

        operations = re.sub(
            "^",
            " " * 4,
            "\n\n".join([str(a) for a in self._sequence]),
            flags=re.MULTILINE,
        )
        operations = re.sub("^\\s+$", "", operations, flags=re.MULTILINE)

        return (
            "LUT Sequence\n"
            "------------\n\n"
            "Overview\n\n"
            f"    {sequence}\n\n"
            "Operations\n\n"
            f"{operations}"
        )

    def __repr__(self) -> str:
        """
        Return an evaluable string representation of the *LUT* sequence.

        Returns
        -------
        :class:`str`
            Evaluable string representation.
        """

        operations = re.sub(
            "^",
            " " * 4,
            ",\n".join([repr(a) for a in self._sequence]),
            flags=re.MULTILINE,
        )
        operations = re.sub("^\\s+$", "", operations, flags=re.MULTILINE)

        return f"{self.__class__.__name__}(\n{operations}\n)"

    def __eq__(self, other) -> bool:
        """
        Return whether the *LUT* sequence is equal to given other object.

        Parameters
        ----------
        other
            Object to test whether it is equal to the *LUT* sequence.

        Returns
        -------
        :class:`bool`
            Whether given object is equal to the *LUT* sequence.
        """

        if not isinstance(other, LUTSequence):
            return False

        if len(self) != len(other):
            return False

        # pylint: disable=C0200
        for i in range(len(self)):
            if self[i] != other[i]:
                return False

        return True

    def __ne__(self, other) -> bool:
        """
        Return whether the *LUT* sequence is not equal to given other object.

        Parameters
        ----------
        other
            Object to test whether it is not equal to the *LUT* sequence.

        Returns
        -------
        :class:`bool`
            Whether given object is not equal to the *LUT* sequence.
        """

        return not (self == other)

    def insert(self, index: Integer, item: TypeLUTSequenceItem):
        """
        Insert given *LUT* at given index into the *LUT* sequence.

        Parameters
        ----------
        index
            Index to insert the item at into the *LUT* sequence.
        item
            *LUT* to insert into the *LUT* sequence.
        """

        attest(
            isinstance(item, TypeLUTSequenceItem),
            '"value" items must implement the "TypeLUTSequenceItem" '
            "protocol!",
        )

        self._sequence.insert(index, item)

    def apply(self, RGB: ArrayLike, **kwargs: Any) -> NDArray:
        """
        Apply the *LUT* sequence sequentially to given *RGB* colourspace
        array.

        Parameters
        ----------
        RGB
            *RGB* colourspace array to apply the *LUT* sequence sequentially
            onto.

        Other Parameters
        ----------------
        kwargs
            Keywords arguments, the keys must be the class type names for which
            they are intended to be used with. There is no implemented way to
            discriminate which class instance the keyword arguments should be
            used with, thus if many class instances of the same type are
            members of the sequence, any matching keyword arguments will be
            used with all the class instances.

        Returns
        -------
        :class:`numpy.ndarray`
            Processed *RGB* colourspace array.

        Examples
        --------
        >>> import numpy as np
        >>> from colour.io.luts import  LUT1D, LUT3x1D, LUT3D
        >>> from colour.utilities import tstack
        >>> LUT_1 = LUT1D(LUT1D.linear_table(16) + 0.125)
        >>> LUT_2 = LUT3D(LUT3D.linear_table(16) ** (1 / 2.2))
        >>> LUT_3 = LUT3x1D(LUT3x1D.linear_table(16) * 0.750)
        >>> LUT_sequence = LUTSequence(LUT_1, LUT_2, LUT_3)
        >>> samples = np.linspace(0, 1, 5)
        >>> RGB = tstack([samples, samples, samples])
        >>> LUT_sequence.apply(RGB, LUT1D={'direction': 'Inverse'})
        ... # doctest: +ELLIPSIS
        array([[ 0.       ...,  0.       ...,  0.       ...],
               [ 0.2899886...,  0.2899886...,  0.2899886...],
               [ 0.4797662...,  0.4797662...,  0.4797662...],
               [ 0.6055328...,  0.6055328...,  0.6055328...],
               [ 0.7057779...,  0.7057779...,  0.7057779...]])
        """

        RGB = as_float_array(RGB)

        RGB_o = RGB
        for operator in self:
            RGB_o = operator.apply(
                RGB_o, **kwargs.get(operator.__class__.__name__, {})
            )

        return RGB_o

    def copy(self) -> LUTSequence:
        """
        Return a copy of the *LUT* sequence.

        Returns
        -------
        :class:`colour.LUTSequence`
            *LUT* sequence copy.
        """

        return deepcopy(self)



"""Showcases *Academy Spectral Similarity Index* (SSI) computations."""

import colour
from colour.utilities import message_box

message_box("Academy Spectral Similarity Index Computations")

message_box(
    'Computing the "CIE Illuminant B" "Academy Spectral Similarity Index (SSI)" '
    'with "CIE Standard Illuminant D65".'
)
print(
    colour.spectral_similarity_index(
        colour.SDS_ILLUMINANTS["B"], colour.SDS_ILLUMINANTS["D65"]
    )
)
